"""Tests for the carddav_object.YAMLEditable class"""

import unittest

from .helpers import TestYAMLEditable


class ToYamlConversion(unittest.TestCase):

    def test_yaml_quoted_special_characters(self):
        yaml_editable = TestYAMLEditable()
        yaml_editable.supported_private_objects = ["Twitter"]
        yaml_repr = """
Formatted name: Test vCard
First name: Khard
Private       :
    Twitter: \"@khard\"
"""
        yaml_editable.update(yaml_repr)
        yaml_dump = yaml_editable.to_yaml()
        self.assertIn("'@khard'", yaml_dump)


class ExceptionHandling(unittest.TestCase):

    def test_duplicate_key_errors_are_translated_to_value_errors(self):
        ye = TestYAMLEditable()
        with self.assertRaises(ValueError):
            ye.update("{key: value, key: again}")

    def test_parser_error_is_translated_to_value_error(self):
        ye = TestYAMLEditable()
        with self.assertRaises(ValueError):
            ye.update("{[invalid yaml")



"""Unittests for the khard module"""

from argparse import Namespace
from email.headerregistry import Address
import unittest
from unittest import mock

from khard import khard, query, config
from khard.khard import find_email_addresses

from .helpers import TmpAbook, load_contact


class TestSearchQueryPreparation(unittest.TestCase):

    foo = query.TermQuery("foo")
    bar = query.TermQuery("bar")

    def setUp(self):
        # Set the uninitialized global variable in the khard module to make it
        # mockable. See https://stackoverflow.com/questions/61193676
        khard.config = mock.Mock()

    @staticmethod
    def _make_abook(name):
        abook = mock.Mock()
        abook.name = name
        return abook

    @classmethod
    def _run(cls, **kwargs):
        with mock.patch("khard.khard.config.abooks",
                        [cls._make_abook(name)
                         for name in ["foo", "bar", "baz"]]):
            return khard.prepare_search_queries(Namespace(**kwargs))

    def test_queries_for_the_same_address_book_are_joind_by_disjunction(self):
        expected = self.foo | self.bar
        prepared = self._run(addressbook=["foo"], target_addressbook=["foo"],
                             source_search_terms=self.foo,
                             target_contact=self.bar)
        self.assertEqual(expected, prepared["foo"])

    def test_no_search_terms_result_in_any_queries(self):
        expected = query.AnyQuery()
        prepared = self._run(addressbook=["foo"], target_addressbook=["foo"],
                             source_search_terms=query.AnyQuery(),
                             target_contact=query.AnyQuery())
        self.assertEqual(expected, prepared["foo"])


class TestAddEmail(unittest.TestCase):

    def test_find_email_addresses_empty_text_finds_none(self):
        text = ""
        addrs = find_email_addresses(text, ["from"])
        self.assertEqual([], addrs)

    def test_find_email_addresses_single_header_finds_one_address(self):
        text = """From: John Doe <jdoe@machine.example>"""
        addrs = find_email_addresses(text, ["from"])
        expected = [Address(display_name="John Doe",
                            username="jdoe", domain="machine.example")]
        self.assertEqual(expected, addrs)

    def test_find_email_addresses_single_header_finds_multiple_addresses(self):
        text = """From: John Doe <jdoe@machine.example>, \
                Mary Smith <mary@example.net>"""
        addrs = find_email_addresses(text, ["from"])
        expected = [
            Address(
                display_name="John Doe",
                username="jdoe",
                domain="machine.example"),
            Address(
                display_name="Mary Smith",
                username="mary",
                domain="example.net")]
        self.assertEqual(expected, addrs)

    def test_find_email_addresses_non_address_header_finds_none(self):
        text = "From: John Doe <jdoe@machine.example>, " \
            "Mary Smith <mary@example.net>\nOther: test"
        addrs = find_email_addresses(text, ["other"])
        expected = []
        self.assertEqual(expected, addrs)

    def test_find_email_addresses_multiple_headers_finds_some(self):
        text = "From: John Doe <jdoe@machine.example>, " \
            "Mary Smith <mary@example.net>\nOther: test"
        addrs = find_email_addresses(text, ["other", "from"])
        expected = [
            Address(
                display_name="John Doe",
                username="jdoe",
                domain="machine.example"),
            Address(
                display_name="Mary Smith",
                username="mary",
                domain="example.net")]
        self.assertEqual(expected, addrs)

    def test_find_email_addresses_multiple_headers_finds_all(self):
        text = "From: John Doe <jdoe@machine.example>, " \
            "Mary Smith <mary@example.net>\n" \
            "To: Michael Jones <mjones@machine.example>"
        addrs = find_email_addresses(text, ["to", "FrOm"])
        expected = [
            Address(
                display_name="Michael Jones",
                username="mjones",
                domain="machine.example"),
            Address(
                display_name="John Doe",
                username="jdoe",
                domain="machine.example"),
            Address(
                display_name="Mary Smith",
                username="mary",
                domain="example.net")]
        self.assertEqual(expected, addrs)

    def test_find_email_addresses_finds_all_emails(self):
        text = "From: John Doe <jdoe@machine.example>, " \
            "Mary Smith <mary@example.net>\n" \
            "To: Michael Jones <mjones@machine.example>"
        addrs = find_email_addresses(text, ["all"])
        expected = [
            Address(
                display_name="John Doe",
                username="jdoe",
                domain="machine.example"),
            Address(
                display_name="Mary Smith",
                username="mary",
                domain="example.net"),
            Address(
                display_name="Michael Jones",
                username="mjones",
                domain="machine.example")]
        self.assertEqual(expected, addrs)

    def test_find_email_addresses_finds_all_emails_with_other_headers_too(
            self):
        text = "From: John Doe <jdoe@machine.example>, " \
            "Mary Smith <mary@example.net>\n" \
            "To: Michael Jones <mjones@machine.example>"
        addrs = find_email_addresses(text, ["other", "all", "from"])
        expected = [
            Address(
                display_name="John Doe",
                username="jdoe",
                domain="machine.example"),
            Address(
                display_name="Mary Smith",
                username="mary",
                domain="example.net"),
            Address(
                display_name="Michael Jones",
                username="mjones",
                domain="machine.example")]
        self.assertEqual(expected, addrs)


class TestGetContactListByUserSelection(unittest.TestCase):

    def setUp(self):
        """initialize the global config object with a mock"""
        khard.config = mock.Mock(spec=config.Config)
        khard.config.group_by_addressbook = False
        khard.config.reverse = False
        khard.config.sort = "last_name"

    def tearDown(self):
        del khard.config

    def test_uid_query_without_strict_search(self):
        q = query.FieldQuery("uid", "testuid1")
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            l = khard.get_contact_list(abook, q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, 'testuid1')

    def test_name_query_with_uid_text_and_strict_search(self):
        q = query.NameQuery("testuid1")
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            l = khard.get_contact_list(abook, q)
        self.assertEqual(len(l), 0)

    def test_name_query_with_uid_text_and_without_strict_search(self):
        q = query.NameQuery("testuid1")
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            l = khard.get_contact_list(abook, q)
        self.assertEqual(len(l), 0)

    def test_term_query_without_strict_search(self):
        q = query.TermQuery("testuid1")
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            l = khard.get_contact_list(abook, q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, 'testuid1')

    def test_term_query_with_strict_search_matching(self):
        q = query.TermQuery("second contact")
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            l = khard.get_contact_list(abook, q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, 'testuid1')


class TestSortContacts(unittest.TestCase):

    contact1 = load_contact("contact1.vcf")
    contact2 = load_contact("contact2.vcf")
    nickname = load_contact("nickname.vcf")
    no_nickname = load_contact("no-nickname.vcf")

    def _test(self, first, second, **kwargs):
        """Run the sort_contacts function and assert the result

        The two contacts first and second are expected to come out in that
        order and are deliberatly put into the function in the reverse order.
        """
        actual = khard.sort_contacts([second, first], **kwargs)
        self.assertListEqual(actual, [first, second])

    def test_sorts_by_first_name_by_default(self):
        self._test(self.nickname, self.no_nickname)

    def test_reverses_sort_order(self):
        self._test(self.no_nickname, self.nickname, reverse=True)

    def test_can_sort_by_last_name(self):
        self._test(self.no_nickname, self.nickname, sort="last_name")

    def test_can_sort_by_formatted_name(self):
        self._test(self.contact1, self.contact2, sort="formatted_name")

    def test_group_by_addressbook(self):
        with TmpAbook(["contact1.vcf", "category.vcf"], name="one") as abook1:
            with TmpAbook(["contact2.vcf", "labels.vcf"],
                          name="two") as abook2:
                contact1 = next(abook1.search(query.FieldQuery("uid",
                                                               "testuid1")))
                category = next(abook1.search(query.NameQuery("category")))
                contact2 = next(abook2.search(query.FieldQuery("uid",
                                                               "testuid2")))
                labels = next(abook2.search(query.NameQuery("labeled guy")))
        expected = [category, contact1, labels, contact2]
        actual = khard.sort_contacts([contact1, contact2, category, labels],
                                     group=True)
        self.assertListEqual(actual, expected)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os

# twinkle config folder
twinkle_config = os.path.join(os.environ['HOME'], ".twinkle")

# khard executable
khard_exe = os.path.join(os.environ['HOME'], ".virtualenvs", "bin", "khard")

# user language
language = "de"

# stop mpd
stop_music = True
mpd_host = "192.168.2.100"
mpd_port = 6600

# log file for calls
call_log_file = os.path.join(twinkle_config, "calls.log")

# audio files
constant_ringtone_segment = os.path.join(twinkle_config, "sounds", "ringtone_segment.wav")
new_ringtone = os.path.join(twinkle_config, "sounds", "special_ringtone.wav")

# temp files
tmp_mono_file = "/tmp/caller_id.wav"
tmp_file_stereo = "/tmp/caller_id_stereo.wav"
mpd_lockfile = "/tmp/mpd_stopped"
caller_id_filename = "/tmp/current_caller_id"




"""Helper functions for the tests."""
# pylint: disable=invalid-name

import contextlib
import io
import os
import shutil
import tempfile
from unittest import mock

import vobject

from khard import address_book
from khard import carddav_object


def vCard(**kwargs):
    """Create a simple vobject.vCard for tests."""
    vcard = vobject.vCard()
    if 'fn' not in kwargs:
        kwargs['fn'] = 'Test vCard'
    if 'version' not in kwargs:
        kwargs['version'] = '3.0'
    for key, value in kwargs.items():
        vcard.add(key.upper()).value = value
    return vcard


def TestVCardWrapper(**kwargs):
    """Create a simple VCardWrapper for tests."""
    return carddav_object.VCardWrapper(vCard(**kwargs))


def TestYAMLEditable(**kwargs):
    """Create a simple YAMLEditable for tests."""
    return carddav_object.YAMLEditable(vCard(**kwargs))


def TestCarddavObject(**kwargs):
    """Create a siple CarddavObject for tests."""
    return carddav_object.CarddavObject(vCard(**kwargs), None, None)


def mock_stream(name="stdout"):
    """A context manager to replace a stdio stream with a string buffer.

    >>> with mock_stream() as s:
    >>>     print("hello world")
    >>> assert s.getvalue() == "hello world"
    >>> with mock_stream("stderr") as e:
    >>>     print("hallo error", file=sys.stderr)
    >>> assert e.getvalue() == "hello error"
    """
    stream = io.StringIO()
    context_manager = mock.patch('sys.'+name, stream)
    context_manager.getvalue = stream.getvalue
    return context_manager


def load_contact(path, abook=None):
    """Load a contact from the fixture directory.

    :param str path: the file name (full, relative to cwd or the fixture dir)
    :param AddressBook abook:
    :returns CarddavObject:
    """
    if not os.path.exists(path):
        path = os.path.join("test/fixture/vcards", path)
    return carddav_object.CarddavObject.from_file(abook, path)


class TmpAbook:
    """Context manager to create a temporary address book folder"""

    def __init__(self, vcards, name="tmp"):
        self.vcards = vcards
        self.name = name

    def __enter__(self):
        self.tempdir = tempfile.TemporaryDirectory()
        for card in self.vcards:
            shutil.copy(self._card_path(card), self.tempdir.name)
        return address_book.VdirAddressBook(self.name, self.tempdir.name)

    def __exit__(self, _a, _b, _c):
        self.tempdir.cleanup()

    @staticmethod
    def _card_path(card):
        if os.path.exists(card):
            return card
        return os.path.join("test/fixture/vcards", card)


class TmpConfig(contextlib.ContextDecorator):
    """Context manager to create a temporary khard configuration.

    The given vcards will be copied to the only address book in the
    configuration which will be called "tmp".
    """

    def __init__(self, vcards):
        self.tempdir = None
        self.config = None
        self.vcards = vcards
        self.mock = None

    def __enter__(self):
        self.tempdir = tempfile.TemporaryDirectory()
        for card in self.vcards:
            shutil.copy(self._card_path(card), self.tempdir.name)
        with tempfile.NamedTemporaryFile("w", delete=False) as config:
            config.write("""[general]
                            editor = editor
                            merge_editor = merge_editor
                            [addressbooks]
                            [[tmp]]
                            path = {}
                            """.format(self.tempdir.name))
        self.config = config
        self.mock = mock.patch.dict('os.environ', KHARD_CONFIG=config.name)
        self.mock.start()
        return self

    def __exit__(self, _a, _b, _c):
        self.mock.stop()
        os.unlink(self.config.name)
        self.tempdir.cleanup()

    @staticmethod
    def _card_path(card):
        if os.path.exists(card):
            return card
        return os.path.join("test/fixture/vcards", card)



"""Test some features of the command line interface of khard.

This also contains some "end to end" tests.  That means some very high level
calls to the main function and a check against the output.  These might later
be converted to proper "unit" tests.
"""
# pylint: disable=missing-docstring

# TODO We are still missing high level tests for the add-email and merge
# subcommands.  They depend heavily on user interaction and are hard to test in
# their current form.

import io
import pathlib
import shutil
import tempfile
import unittest
from unittest import mock

from ruamel.yaml import YAML

from khard import cli
from khard import config
from khard.helpers.interactive import EditState, Editor
from khard import khard

from .helpers import TmpConfig, mock_stream


def run_main(*args):
    """Run the khard.main() method with mocked stdout"""
    with mock_stream() as stdout:
        khard.main(args)
    return stdout


@mock.patch('sys.argv', ['TESTSUITE'])
class HelpOption(unittest.TestCase):

    def _test(self, args, expect):
        """Test the command line args and compare the prefix of the output."""
        with self.assertRaises(SystemExit):
            with mock_stream() as stdout:
                cli.parse_args(args)
        text = stdout.getvalue()
        self.assertRegex(text, expect)

    def test_global_help(self):
        self._test(['-h'], r'^usage: TESTSUITE \[-h\]')

    @mock.patch.dict('os.environ', KHARD_CONFIG='test/fixture/minimal.conf')
    def test_subcommand_help(self):
        self._test(['list', '-h'], r'^usage: TESTSUITE list \[-h\]')

    def test_global_help_with_subcommand(self):
        self._test(['-h', 'list'], r'^usage: TESTSUITE \[-h\]')


@mock.patch.dict('os.environ', KHARD_CONFIG='test/fixture/minimal.conf')
class ListingCommands(unittest.TestCase):
    """Tests for subcommands that simply list stuff."""


    def test_simple_ls_without_options(self):
        stdout = run_main("list")
        text = [l.strip() for l in stdout.getvalue().splitlines()]
        expected = [
            "Address book: foo",
            "Index    Name              Phone                "
            "Email                     Uid",
            "1        second contact    voice: 0123456789    "
            "home: user@example.com    testuid1",
            "2        text birthday                          "
            "                          testuid3",
            "3        third contact                          "
            "                          testuid2"]
        self.assertListEqual(text, expected)

    def test_ls_fields_like_email(self):
        stdout = run_main('ls', '-p', '-F', 'emails.home.0,name')
        text = stdout.getvalue().splitlines()
        expected = [
            "user@example.com\tsecond contact",
            "\ttext birthday",
            "\tthird contact",
        ]
        self.assertListEqual(text, expected)

    @mock.patch.dict('os.environ', LC_ALL='C')
    def test_simple_bdays_without_options(self):
        stdout = run_main('birthdays')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = ["Name              Birthday",
                  "text birthday     circa 1800",
                  "second contact    01/20/18"]
        self.assertListEqual(text, expect)

    def test_parsable_bdays(self):
        stdout = run_main('birthdays', '--parsable')
        text = stdout.getvalue().splitlines()
        expect = ["circa 1800\ttext birthday", "2018.01.20\tsecond contact"]
        self.assertListEqual(text, expect)

    def test_simple_email_without_options(self):
        stdout = run_main('email')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = ["Name              Type    E-Mail",
                  "second contact    home    user@example.com"]
        self.assertListEqual(text, expect)

    def test_simple_phone_without_options(self):
        stdout = run_main('phone')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = ["Name              Type     Phone",
                  "second contact    voice    0123456789"]
        self.assertListEqual(text, expect)

    def test_simple_file_without_options(self):
        stdout = run_main('filename')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = ["test/fixture/test.abook/contact1.vcf",
                  "test/fixture/test.abook/text-bday.vcf",
                  "test/fixture/test.abook/contact2.vcf"]
        self.assertListEqual(text, expect)

    def test_simple_abooks_without_options(self):
        stdout = run_main('addressbooks')
        text = stdout.getvalue().strip()
        expect = "foo"
        self.assertEqual(text, expect)

    def test_simple_details_without_options(self):
        stdout = run_main('details', 'uid1')
        text = stdout.getvalue()
        # Currently the FN field is not shown with "details".
        self.assertIn('Address book: foo', text)
        self.assertIn('UID: testuid1', text)

    def test_order_of_search_term_does_not_matter(self):
        stdout1 = run_main('list', 'second', 'contact')
        stdout2 = run_main('list', 'contact', 'second')
        text1 = [l.strip() for l in stdout1.getvalue().splitlines()]
        text2 = [l.strip() for l in stdout2.getvalue().splitlines()]
        expected = [
            "Address book: foo",
            "Index    Name              Phone                "
            "Email                     Uid",
            "1        second contact    voice: 0123456789    "
            "home: user@example.com    testuid1"]
        self.assertListEqual(text1, expected)
        self.assertListEqual(text2, expected)

    def test_case_of_search_terms_does_not_matter(self):
        stdout1 = run_main('list', 'second', 'contact')
        stdout2 = run_main('list', 'SECOND', 'CONTACT')
        text1 = [l.strip() for l in stdout1.getvalue().splitlines()]
        text2 = [l.strip() for l in stdout2.getvalue().splitlines()]
        expected = [
            "Address book: foo",
            "Index    Name              Phone                "
            "Email                     Uid",
            "1        second contact    voice: 0123456789    "
            "home: user@example.com    testuid1"]
        self.assertListEqual(text1, expected)
        self.assertListEqual(text2, expected)

    def test_regex_special_chars_are_not_special(self):
        with self.assertRaises(SystemExit):
            with mock_stream() as stdout:
                khard.main(['list', 'uid.'])
        self.assertEqual(stdout.getvalue(), "Found no contacts\n")

    def test_display_post_address(self):
        with TmpConfig(["post.vcf"]):
            stdout = run_main('postaddress')
        text = [line.rstrip() for line in stdout.getvalue().splitlines()]
        expected = [
            'Name                 Type    Post address',
            'With post address    home    Main Street 1',
            '                             PostBox Ext',
            '                             00000 The City',
            '                             SomeState, HomeCountry']

        self.assertListEqual(expected, text)

    def test_email_lists_only_contacts_with_emails(self):
        with TmpConfig(["contact1.vcf", "contact2.vcf"]):
            stdout = run_main("email")
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = ["Name              Type    E-Mail",
                  "second contact    home    user@example.com"]
        self.assertListEqual(expect, text)

    def test_phone_lists_only_contacts_with_phone_nubers(self):
        with TmpConfig(["contact1.vcf", "contact2.vcf"]):
            stdout = run_main("phone")
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = ["Name              Type     Phone",
                  "second contact    voice    0123456789"]
        self.assertListEqual(expect, text)

    def test_postaddr_lists_only_contacts_with_post_addresses(self):
        with TmpConfig(["contact1.vcf", "post.vcf"]):
            stdout = run_main("postaddress")
        text = [line.rstrip() for line in stdout.getvalue().splitlines()]
        expect = ['Name                 Type    Post address',
                  'With post address    home    Main Street 1',
                  '                             PostBox Ext',
                  '                             00000 The City',
                  '                             SomeState, HomeCountry']
        self.assertListEqual(expect, text)

    def test_mixed_kinds(self):
        with TmpConfig(["org.vcf", "individual.vcf"]):
            stdout = run_main("list", "organisations:acme")
        text = [line.rstrip() for line in stdout.getvalue().splitlines()]
        expected = [
            "Address book: tmp",
            "Index    Name              Phone    Email    Kind            Uid",
            "1        ACME Inc.                           organisation    4",
            "2        Wile E. Coyote                      individual      1"]
        self.assertListEqual(expected, text)

    def test_non_individual_kind(self):
        with TmpConfig(["org.vcf"]):
            stdout = run_main("list")
        text = [line.rstrip() for line in stdout.getvalue().splitlines()]
        expected = [
            "Address book: tmp",
            "Index    Name         Phone    Email    Kind            Uid",
            "1        ACME Inc.                      organisation    4"]
        self.assertListEqual(expected, text)


class ListingCommands2(unittest.TestCase):

    def test_list_bug_195(self):
        with TmpConfig(['tel-value-uri.vcf']):
            stdout = run_main('list')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = [
            "Address book: tmp",
            "Index    Name       Phone             Email    Uid",
            "1        bug 195    cell: 67545678             b"]
        self.assertListEqual(text, expect)

    def test_list_bug_243_part_1(self):
        """Search for a category with the ls command"""
        with TmpConfig(['category.vcf']):
            stdout = run_main('list', 'bar')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = [
            "Address book: tmp",
            "Index    Name                     Phone    "
            "Email                        Uid",
            "1        contact with category             "
            "internet: foo@example.org    c",
        ]
        self.assertListEqual(text, expect)

    def test_list_bug_243_part_2(self):
        """Search for a category with the email command"""
        with TmpConfig(['category.vcf']):
            stdout = run_main('email', 'bar')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = [
            "Name                     Type        E-Mail",
            "contact with category    internet    foo@example.org",
        ]
        self.assertListEqual(text, expect)

    def test_list_bug_251(self):
        "Find contacts by nickname even if a match by name exists"
        with TmpConfig(["test/fixture/nick.abook/nickname.vcf",
                        "test/fixture/vcards/no-nickname.vcf"]):
            stdout = run_main('list', 'mike')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = ['Address book: tmp',
                  'Index    Name             Phone    Email                   '
                  'Uid',
                  '1        Michael Smith             pref: ms@example.org    '
                  'issue251part1',
                  '2        Mike Jones                pref: mj@example.org    '
                  'issue251part2']
        self.assertListEqual(text, expect)

    @mock.patch.dict('os.environ', KHARD_CONFIG='test/fixture/nick.conf')
    def test_email_bug_251(self):
        stdout = run_main('email', '--parsable', 'mike')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = ["searching for 'mike' ...",
                  "ms@example.org\tMichael Smith\tpref"]
        self.assertListEqual(text, expect)

    @mock.patch.dict('os.environ', KHARD_CONFIG='test/fixture/nick.conf')
    def test_email_bug_251_part2(self):
        stdout = run_main('email', '--parsable', 'joe')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = ["searching for 'joe' ...",
                  "jcitizen@foo.com\tJoe Citizen\tpref"]
        self.assertListEqual(text, expect)

    def test_email_bug_251_part_3(self):
        "Find contacts by nickname even if a match by name exists"
        with TmpConfig(["test/fixture/nick.abook/nickname.vcf",
                        "test/fixture/vcards/no-nickname.vcf"]):
            stdout = run_main('email', '--parsable', 'mike')
        text = [line.strip() for line in stdout.getvalue().splitlines()]
        expect = ["searching for 'mike' ...",
                  'ms@example.org\tMichael Smith\tpref',
                  'mj@example.org\tMike Jones\tpref']
        self.assertListEqual(text, expect)


class FileSystemCommands(unittest.TestCase):
    """Tests for subcommands that interact with different address books."""

    def setUp(self):
        "Create a temporary directory with two address books and a configfile."
        self._tmp = tempfile.TemporaryDirectory()
        path = pathlib.Path(self._tmp.name)
        self.abook1 = path / 'abook1'
        self.abook2 = path / 'abook2'
        self.abook1.mkdir()
        self.abook2.mkdir()
        self.contact = self.abook1 / 'contact.vcf'
        shutil.copy('test/fixture/vcards/contact1.vcf', str(self.contact))
        config = path / 'conf'
        with config.open('w') as fh:
            fh.write("""[addressbooks]
                        [[abook1]]
                        path = {}
                        [[abook2]]
                        path = {}""".format(self.abook1, self.abook2))
        self._patch = mock.patch.dict('os.environ', KHARD_CONFIG=str(config))
        self._patch.start()

    def tearDown(self):
        self._patch.stop()
        self._tmp.cleanup()

    def test_simple_move(self):
        # just hide stdout
        with mock.patch('sys.stdout'):
            khard.main(['move', '-a', 'abook1', '-A', 'abook2', 'testuid1'])
        # The contact is moved to a filename based on the uid.
        target = self.abook2 / 'testuid1.vcf'
        # We currently only assert that the target file exists, nothing about
        # its contents.
        self.assertFalse(self.contact.exists())
        self.assertTrue(target.exists())

    def test_simple_copy(self):
        # just hide stdout
        with mock.patch('sys.stdout'):
            khard.main(['copy', '-a', 'abook1', '-A', 'abook2', 'testuid1'])
        # The contact is copied to a filename based on a new uid.
        results = list(self.abook2.glob('*.vcf'))
        self.assertTrue(self.contact.exists())
        self.assertEqual(len(results), 1)

    def test_simple_remove_with_force_option(self):
        # just hide stdout
        with mock.patch('sys.stdout'):
            # Without the --force this asks for confirmation.
            khard.main(['remove', '--force', '-a', 'abook1', 'testuid1'])
        results = list(self.abook2.glob('*.vcf'))
        self.assertFalse(self.contact.exists())
        self.assertEqual(len(results), 0)

    def test_new_contact_with_simple_user_input(self):
        old = len(list(self.abook1.glob('*.vcf')))
        # Mock user input on stdin (yaml format).
        with mock.patch('sys.stdin.isatty', return_value=False):
            with mock.patch('sys.stdin.read',
                            return_value='First name: foo\nLast name: bar'):
                # just hide stdout
                with mock.patch('sys.stdout'):
                    # hide warning about missing version in vcard
                    with self.assertLogs(level='WARNING'):
                        khard.main(['new', '-a', 'abook1'])
        new = len(list(self.abook1.glob('*.vcf')))
        self.assertEqual(new, old + 1)


class MiscCommands(unittest.TestCase):
    """Tests for other subcommands."""

    @mock.patch.dict('os.environ', KHARD_CONFIG='test/fixture/minimal.conf')
    def test_simple_show_with_yaml_format(self):
        stdout = run_main("show", "--format=yaml", "uid1")
        # This implicitly tests if the output is valid yaml.
        yaml = YAML(typ="base").load(stdout.getvalue())
        # Just test some keys.
        self.assertIn('Address', yaml)
        self.assertIn('Birthday', yaml)
        self.assertIn('Email', yaml)
        self.assertIn('First name', yaml)
        self.assertIn('Last name', yaml)
        self.assertIn('Nickname', yaml)

    @mock.patch.dict('os.environ', KHARD_CONFIG='test/fixture/minimal.conf')
    def test_simple_edit_without_modification(self):
        editor = mock.Mock()
        editor.edit_templates = mock.Mock(return_value=None)
        editor.write_temp_file = Editor.write_temp_file
        with mock.patch('khard.khard.interactive.Editor',
                        mock.Mock(return_value=editor)):
            run_main("edit", "uid1")
        # The editor is called with a temp file so how to we check this more
        # precisely?
        editor.edit_templates.assert_called_once()

    @mock.patch.dict('os.environ', KHARD_CONFIG='test/fixture/minimal.conf',
                     EDITOR='editor')
    def test_edit_source_file_without_modifications(self):
        with mock.patch('subprocess.Popen') as popen:
            run_main("edit", "--format=vcard", "uid1")
        popen.assert_called_once_with(['editor',
                                       'test/fixture/test.abook/contact1.vcf'])


@mock.patch.dict('os.environ', KHARD_CONFIG='test/fixture/minimal.conf')
class CommandLineDefaultsDoNotOverwriteConfigValues(unittest.TestCase):

    @staticmethod
    def _with_contact_table(args, **kwargs):
        args = cli.parse_args(args)
        options = '\n'.join('{}={}'.format(key, kwargs[key]) for key in kwargs)
        conf = config.Config(io.StringIO('[addressbooks]\n[[test]]\npath=.\n'
                                         '[contact table]\n' + options))
        return cli.merge_args_into_config(args, conf)

    def test_group_by_addressbook(self):
        conf = self._with_contact_table(['list'], group_by_addressbook=True)
        self.assertTrue(conf.group_by_addressbook)


@mock.patch.dict('os.environ', KHARD_CONFIG='test/fixture/minimal.conf')
class CommandLineArgumentsOverwriteConfigValues(unittest.TestCase):

    @staticmethod
    def _merge(args):
        args, _conf = cli.parse_args(args)
        # This config file just loads all defaults from the config.spec.
        conf = config.Config(io.StringIO('[addressbooks]\n[[test]]\npath=.'))
        return cli.merge_args_into_config(args, conf)

    def test_sort_is_picked_up_from_arguments(self):
        conf = self._merge(['list', '--sort=last_name'])
        self.assertEqual(conf.sort, 'last_name')

    def test_display_is_picked_up_from_arguments(self):
        conf = self._merge(['list', '--display=last_name'])
        self.assertEqual(conf.display, 'last_name')

    def test_reverse_is_picked_up_from_arguments(self):
        conf = self._merge(['list', '--reverse'])
        self.assertTrue(conf.reverse)

    def test_group_by_addressbook_is_picked_up_from_arguments(self):
        conf = self._merge(['list', '--group-by-addressbook'])
        self.assertTrue(conf.group_by_addressbook)

    def test_search_in_source_is_picked_up_from_arguments(self):
        conf = self._merge(['list', '--search-in-source-files'])
        self.assertTrue(conf.search_in_source_files)


class Merge(unittest.TestCase):

    def test_merge_with_exact_search_terms(self):
        with TmpConfig(["contact1.vcf", "contact2.vcf"]):
            with mock.patch('khard.khard.merge_existing_contacts') as merge:
                run_main("merge", "second", "--target", "third")
        merge.assert_called_once()
        # unpack the call arguments
        call = merge.mock_calls[0]
        name, args, kwargs = call
        first, second, delete = args
        self.assertTrue(delete)
        first = pathlib.Path(first.filename).name
        second = pathlib.Path(second.filename).name
        self.assertEqual('contact1.vcf', first)
        self.assertEqual('contact2.vcf', second)

    def test_merge_with_exact_uid_search_terms(self):
        with TmpConfig(["contact1.vcf", "contact2.vcf"]):
            with mock.patch('khard.khard.merge_existing_contacts') as merge:
                run_main("merge", "--uid", "testuid1", "--target-uid",
                         "testuid2")
        merge.assert_called_once()
        # unpack the call arguments
        call = merge.mock_calls[0]
        name, args, kwargs = call
        first, second, delete = args
        self.assertTrue(delete)
        first = pathlib.Path(first.filename).name
        second = pathlib.Path(second.filename).name
        self.assertEqual('contact1.vcf', first)
        self.assertEqual('contact2.vcf', second)


class AddEmail(unittest.TestCase):

    # FIXME the new code from fdc441cf asks for confirmation in
    # khard.add_email_to_contact on line 419
    @unittest.skip("unexpected read from stdin blocks the test")
    @TmpConfig(["contact1.vcf", "contact2.vcf"])
    def test_contact_is_found_if_name_matches(self):
        email = [
            "From: third <third@example.com>\n",
            "To: anybody@example.com\n",
            "\n",
            "text\n"
        ]
        with tempfile.NamedTemporaryFile("w") as tmp:
            tmp.writelines(email)
            tmp.flush()
            with mock.patch("khard.khard.confirm", lambda x: True):
                with mock.patch("builtins.input", lambda x: ""):
                    run_main("add-email", "--input-file", tmp.name)
        stdout = run_main("list", "--fields=emails.internet.0")
        addr = stdout.getvalue().splitlines()[-1].strip()
        self.assertEqual(addr, "third@example.com")


if __name__ == "__main__":
    unittest.main()



"""Tests for the vcard formatting functions"""

import unittest

from vobject.vcard import Name

from khard.carddav_object import CarddavObject
from khard.formatter import Formatter

from .helpers import vCard


class FormatLabeledField(unittest.TestCase):

    def test_labels_are_selected_alphabetically_if_no_preferred_given(self):
        labeled_field = {'some': ['thing'], 'other': ['thing']}
        preferred = []
        expected = 'other: thing'
        self.assertEqual(expected, Formatter.format_labeled_field(
            labeled_field, preferred))

    def test_labels_are_selected_alphabetically_if_no_preferred_matches(self):
        labeled_field = {'some': ['thing'], 'other': ['thing']}
        preferred = ['nonexistent']
        expected = 'other: thing'
        self.assertEqual(expected, Formatter.format_labeled_field(
            labeled_field, preferred))

    def test_preferred_labels_are_used(self):
        labeled_field = {'some': ['thing'], 'other': ['thing']}
        preferred = ['some']
        expected = 'some: thing'
        self.assertEqual(expected, Formatter.format_labeled_field(
            labeled_field, preferred))

    def test_alphabetically_first_value_is_used(self):
        labeled_field = {'some': ['thing', 'more']}
        preferred = []
        expected = 'some: more'
        self.assertEqual(expected, Formatter.format_labeled_field(
            labeled_field, preferred))

    def test_not_only_first_char_of_label_is_used(self):
        preferred = []
        labeled_field = {'x-foo': ['foo'], 'x-bar': ['bar']}
        expected = 'x-bar: bar'
        self.assertEqual(expected, Formatter.format_labeled_field(
            labeled_field, preferred))
        expected = 'bar: bar'
        labeled_field = {'foo': ['foo'], 'bar': ['bar']}
        self.assertEqual(expected, Formatter.format_labeled_field(
            labeled_field, preferred))


class GetSpecialField(unittest.TestCase):

    _name = Name(family='Family', given='Given', additional='Additional',
                 prefix='Prefix', suffix='Suffix')
    _vcard = CarddavObject(vCard(fn="Formatted Name", n=_name,
                                 nickname="Nickname"), None, "")

    def _test_name(self, fmt, nick, parsable, expected):
        f = Formatter(fmt, [], [], nick, parsable)
        actual = f.get_special_field(self._vcard, "name")
        self.assertEqual(expected, actual)

    def test_name_formatted_as_first_name_last_name(self):
        self._test_name(Formatter.FIRST, False, False,
                        "Given Additional Family")

    def test_name_formatted_as_first_name_last_name_with_nickname(self):
        self._test_name(Formatter.FIRST, True, False,
                        "Given Additional Family (Nickname: Nickname)")

    def test_name_formatted_as_last_name_first_name(self):
        self._test_name(Formatter.LAST, False, False,
                        "Family, Given Additional")

    def test_name_formatted_as_last_name_first_name_with_nickname(self):
        self._test_name(Formatter.LAST, True, False,
                        "Family, Given Additional (Nickname: Nickname)")

    def test_name_formatted_as_formatted_name(self):
        self._test_name(Formatter.FORMAT, False, False, "Formatted Name")

    def test_name_formatted_as_formatted_name_with_nickname(self):
        self._test_name(Formatter.FORMAT, True, False,
                        "Formatted Name (Nickname: Nickname)")

    def test_parsable_overrides_nickname_with_first_formatting(self):
        self._test_name(Formatter.FIRST, True, True, "Given Additional Family")

    def test_parsable_overrides_nickname_with_last_formatting(self):
        self._test_name(Formatter.LAST, True, True, "Family, Given Additional")

    def test_parsable_overrides_nickname_with_formatted_name(self):
        self._test_name(Formatter.FORMAT, True, True, "Formatted Name")



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import sys
import argparse
from caldavclientlibrary.client.account import CalDAVAccount
from caldavclientlibrary.protocol.url import URL


def main():
    # create the args parser
    parser = argparse.ArgumentParser(
            description="Davcontroller creates, lists and removes caldav "
            "calendars and carddav address books from server")
    parser.add_argument("-v", "--version", action="store_true",
                        help="Get current program version")
    parser.add_argument("-H", "--hostname", default="")
    parser.add_argument("-p", "--port", default="")
    parser.add_argument("-u", "--username", default="")
    parser.add_argument("-P", "--password", default="")
    parser.add_argument("action", nargs="?", default="",
                        help="Actions: new-addressbook, new-calendar, list, "
                        "and remove")
    args = parser.parse_args()

    # version
    if args.version is True:
        print("davcontroller version 0.1")
        sys.exit(0)

    # check the server's parameter
    if args.hostname == "":
        print("Missing host name")
        sys.exit(1)
    if args.port == "":
        print("Missing host port")
        sys.exit(1)
    if args.username == "":
        print("Missing user name")
        sys.exit(1)
    if args.password == "":
        print("Missing password")
        sys.exit(1)

    if args.action == "":
        print("Please specify an action. Possible values are: "
              "new-addressbook, new-calendar, list and remove")
        sys.exit(1)
    elif args.action not in ["new-addressbook", "new-calendar", "list",
                             "remove"]:
        print("The specified action \"%s\" is not supported. Possible values "
              "are: new-addressbook, new-calendar, list and remove" %
              args.action)
        sys.exit(1)

    # try to connect to the caldav server
    account = CalDAVAccount(args.hostname, args.port, ssl=True,
                            user=args.username, pswd=args.password, root="/",
                            principal="")
    if account.getPrincipal() is None:
        print("Error: Connection refused")
        sys.exit(2)

    if args.action in ["list", "remove"]:
        # address books
        print("Available address books")
        addressbook_list = account.getPrincipal().listAddressBooks()
        if addressbook_list.__len__() == 0:
            print("No address books found")
        else:
            for index, addressbook in enumerate(addressbook_list):
                print("%d. %s" % (index+1, addressbook.getDisplayName()))
        print()
        # calendars
        print("Available calendars")
        calendar_list = account.getPrincipal().listCalendars()
        if calendar_list.__len__() == 0:
            print("No calendars found")
        else:
            for index, calendar in enumerate(calendar_list):
                print("%d. %s" % (addressbook_list.__len__() + index + 1,
                      calendar.getDisplayName()))
        item_list = addressbook_list + calendar_list
        if item_list.__len__() == 0:
            sys.exit(2)

        if args.action == "remove":
            print()
            while True:
                input_string = input("Enter Id: ")
                if input_string == "":
                    sys.exit(0)
                try:
                    id = int(input_string)
                    if id > 0 and id <= item_list.__len__():
                        break
                except ValueError:
                    pass
                print("Please enter an Id between 1 and %d or nothing to exit."
                      % item_list.__len__())
            item = item_list[id-1]
            while True:
                input_string = input("Deleting %s. Are you sure? (y/n): " %
                                     item.getDisplayName())
                if input_string.lower() in ["", "n", "q"]:
                    print("Canceled")
                    sys.exit(0)
                if input_string.lower() == "y":
                    break
            account.session.deleteResource(URL(url=item.path))

    if args.action.startswith("new-"):
        # get full host url
        host_url = "https://%s:%s" % (account.session.server,
                                      account.session.port)
        # enter new name
        if args.action == "new-addressbook":
            input_string = input("Enter new address book name or nothing to "
                                 "cancel: ")
        else:
            input_string = input("Enter new calendar name or nothing to "
                                 "cancel: ")
        if input_string == "":
            sys.exit(0)
        res_name = input_string
        res_path = input_string.replace(" ", "_").lower()
        # create new resource
        if args.action == "new-addressbook":
            u = URL(url=host_url + account.principal.adbkhomeset[0].path +
                    res_path + "/")
            account.session.makeAddressBook(u, res_name)
        else:
            u = URL(url=host_url + account.principal.homeset[0].path +
                    res_path + "/")
            account.session.makeCalendar(u, res_name)
        print("Creation successful")


if __name__ == "__main__":
    main()



"""Tests for the VCardWrapper class from the carddav module."""
# pylint: disable=missing-docstring

import datetime
import unittest

import vobject

from khard.carddav_object import VCardWrapper

from .helpers import vCard, TestVCardWrapper


def _from_file(path):
    """Read a VCARD from a file"""
    with open(path) as fp:
        return vobject.readOne(fp)


class VcardWrapperInit(unittest.TestCase):

    def test_stores_vcard_object_unmodified(self):
        vcard = vCard()
        expected = vcard.serialize()
        wrapper = VCardWrapper(vcard)
        # assert that it is the same object
        self.assertIs(wrapper.vcard, vcard)
        # assert that it (the serialization) was not changed
        self.assertEqual(wrapper.vcard.serialize(), expected)

    def test_warns_about_unsupported_version(self):
        with self.assertLogs(level="WARNING"):
            TestVCardWrapper(version="something unsupported")

    def test_warns_about_missing_version_and_sets_it(self):
        vcard = vCard()
        vcard.remove(vcard.version)
        with self.assertLogs(level="WARNING"):
            wrapper = VCardWrapper(vcard)
        self.assertEqual(wrapper.version, "3.0")


class DeleteVcardObject(unittest.TestCase):

    def test_deletes_fields_given_in_upper_case(self):
        vcard = vCard()
        expected = vcard.serialize()
        vcard.add('FOO').value = 'bar'
        wrapper = VCardWrapper(vcard)
        wrapper._delete_vcard_object('FOO')
        self.assertEqual(wrapper.vcard.serialize(), expected)

    def test_deletes_all_field_occurences(self):
        vcard = vCard()
        expected = vcard.serialize()
        vcard.add('FOO').value = 'bar'
        vcard.add('FOO').value = 'baz'
        wrapper = VCardWrapper(vcard)
        wrapper._delete_vcard_object('FOO')
        self.assertEqual(wrapper.vcard.serialize(), expected)

    def test_deletes_grouped_ablabel_fields(self):
        vcard = vCard()
        expected = vcard.serialize()
        foo = vcard.add('FOO')
        foo.value = 'bar'
        foo.group = 'group1'
        label = vcard.add('X-ABLABEL')
        label.value = 'test label'
        label.group = foo.group
        wrapper = VCardWrapper(vcard)
        wrapper._delete_vcard_object('FOO')
        self.assertEqual(wrapper.vcard.serialize(), expected)

    def test_keeps_other_fields(self):
        vcard = vCard(foo='bar')
        expected = vcard.serialize()
        vcard.add('BAR').value = 'baz'
        wrapper = VCardWrapper(vcard)
        wrapper._delete_vcard_object('BAR')
        self.assertEqual(wrapper.vcard.serialize(), expected)

    def test_does_not_fail_on_non_existing_field_name(self):
        vcard = vCard(foo='bar')
        expected = vcard.serialize()
        wrapper = VCardWrapper(vcard)
        wrapper._delete_vcard_object('BAR')
        self.assertEqual(wrapper.vcard.serialize(), expected)


class BirthdayLikeAttributes(unittest.TestCase):

    def test_birthday_supports_setting_date_objects(self):
        wrapper = TestVCardWrapper()
        date = datetime.datetime(2018, 2, 1)
        wrapper.birthday = date
        wrapper.vcard.validate()
        self.assertEqual(wrapper.birthday, date)

    def test_birthday_supports_setting_datetime_objects(self):
        wrapper = TestVCardWrapper()
        date = datetime.datetime(2018, 2, 1, 19, 29, 31)
        wrapper.birthday = date
        wrapper.vcard.validate()
        self.assertEqual(wrapper.birthday, date)

    def test_birthday_supports_setting_text_values_for_v4(self):
        vcard = vCard(version="4.0")
        wrapper = VCardWrapper(vcard, "4.0")
        date = 'some time yesterday'
        wrapper.birthday = date
        wrapper.vcard.validate()
        self.assertEqual(wrapper.birthday, date)

    def test_birthday_does_not_support_setting_text_values_for_v3(self):
        wrapper = TestVCardWrapper(version="3.0")
        with self.assertLogs(level='WARNING'):
            wrapper.birthday = 'some time yesterday'
        wrapper.vcard.validate()
        self.assertIsNone(wrapper.birthday)

    def test_anniversary_supports_setting_date_objects(self):
        wrapper = TestVCardWrapper()
        date = datetime.datetime(2018, 2, 1)
        wrapper.anniversary = date
        wrapper.vcard.validate()
        self.assertEqual(wrapper.anniversary, date)

    def test_anniversary_supports_setting_datetime_objects(self):
        wrapper = TestVCardWrapper()
        date = datetime.datetime(2018, 2, 1, 19, 29, 31)
        wrapper.anniversary = date
        wrapper.vcard.validate()
        self.assertEqual(wrapper.anniversary, date)

    def test_anniversary_supports_setting_text_values_for_v4(self):
        vcard = vCard(version="4.0")
        wrapper = VCardWrapper(vcard, "4.0")
        date = 'some time yesterday'
        wrapper.anniversary = date
        wrapper.vcard.validate()
        self.assertEqual(wrapper.anniversary, date)

    def test_anniversary_does_not_support_setting_text_values_for_v3(self):
        wrapper = TestVCardWrapper(version="3.0")
        with self.assertLogs(level='WARNING'):
            wrapper.birthday = 'some time yesterday'
        wrapper.vcard.validate()
        self.assertIsNone(wrapper.anniversary)


class NameAttributes(unittest.TestCase):

    def test_fn_can_be_set_with_a_string(self):
        vcard = vCard()
        wrapper = VCardWrapper(vcard)
        wrapper.formatted_name = 'foo bar'
        self.assertEqual(vcard.fn.value, 'foo bar')

    def test_only_one_fn_will_be_stored(self):
        vcard = vCard()
        wrapper = VCardWrapper(vcard)
        wrapper.formatted_name = 'foo bar'
        self.assertEqual(len(vcard.contents['fn']), 1)

    def test_fn_is_returned_as_string(self):
        wrapper = TestVCardWrapper()
        self.assertIsInstance(wrapper.formatted_name, str)

    def test_fn_is_used_as_string_representation(self):
        wrapper = TestVCardWrapper()
        self.assertEqual(str(wrapper), wrapper.formatted_name)

    def test_name_can_be_set_with_empty_strings(self):
        vcard = vCard()
        wrapper = VCardWrapper(vcard)
        wrapper._add_name('', '', '', '', '')
        self.assertEqual(vcard.serialize(),
                         'BEGIN:VCARD\r\n'
                         'VERSION:3.0\r\n'
                         'FN:Test vCard\r\n'
                         'N:;;;;\r\n'
                         'END:VCARD\r\n')

    def test_name_can_be_set_with_empty_lists(self):
        vcard = vCard()
        wrapper = VCardWrapper(vcard)
        wrapper._add_name([], [], [], [], [])
        self.assertEqual(vcard.serialize(),
                         'BEGIN:VCARD\r\n'
                         'VERSION:3.0\r\n'
                         'FN:Test vCard\r\n'
                         'N:;;;;\r\n'
                         'END:VCARD\r\n')

    def test_name_can_be_set_with_lists_of_empty_strings(self):
        vcard = vCard()
        wrapper = VCardWrapper(vcard)
        wrapper._add_name(['', ''], ['', ''], ['', ''], ['', ''], ['', ''])
        self.assertEqual(vcard.serialize(),
                         'BEGIN:VCARD\r\n'
                         'VERSION:3.0\r\n'
                         'FN:Test vCard\r\n'
                         'N:;;;;\r\n'
                         'END:VCARD\r\n')

    def test_get_first_name_last_name_retunrs_fn_if_no_name_present(self):
        wrapper = TestVCardWrapper()
        self.assertEqual(wrapper.get_first_name_last_name(), 'Test vCard')

    def test_get_first_name_last_name_with_simple_name(self):
        wrapper = TestVCardWrapper()
        wrapper._add_name('', 'given', '', 'family', '')
        self.assertEqual(wrapper.get_first_name_last_name(), "given family")

    def test_get_first_name_last_name_with_all_name_fields(self):
        wrapper = TestVCardWrapper()
        wrapper._add_name('prefix', 'given', 'additional', 'family', 'suffix')
        self.assertEqual(wrapper.get_first_name_last_name(),
                         'given additional family')

    def test_get_first_name_last_name_with_complex_name(self):
        wrapper = TestVCardWrapper()
        wrapper._add_name(['prefix1', 'prefix2'], ['given1', 'given2'],
                          ['additional1', 'additional2'],
                          ['family1', 'family2'], ['suffix1', 'suffix2'])
        self.assertEqual(wrapper.get_first_name_last_name(), 'given1 given2 '
                         'additional1 additional2 family1 family2')

    def test_get_last_name_first_name_retunrs_fn_if_no_name_present(self):
        wrapper = TestVCardWrapper()
        self.assertEqual(wrapper.get_last_name_first_name(), 'Test vCard')

    def test_get_last_name_first_name_with_simple_name(self):
        wrapper = TestVCardWrapper()
        wrapper._add_name('', 'given', '', 'family', '')
        self.assertEqual(wrapper.get_last_name_first_name(), "family, given")

    def test_get_last_name_first_name_with_all_name_fields(self):
        wrapper = TestVCardWrapper()
        wrapper._add_name('prefix', 'given', 'additional', 'family', 'suffix')
        self.assertEqual(wrapper.get_last_name_first_name(),
                         'family, given additional')

    def test_get_last_name_first_name_with_complex_name(self):
        wrapper = TestVCardWrapper()
        wrapper._add_name(['prefix1', 'prefix2'], ['given1', 'given2'],
                          ['additional1', 'additional2'],
                          ['family1', 'family2'], ['suffix1', 'suffix2'])
        self.assertEqual(wrapper.get_last_name_first_name(), 'family1 family2,'
                         ' given1 given2 additional1 additional2')


class TypedProperties(unittest.TestCase):

    def test_adding_a_simple_phone_number(self):
        wrapper = TestVCardWrapper()
        wrapper._add_phone_number('home', '0123456789')
        self.assertDictEqual(wrapper.phone_numbers, {'home': ['0123456789']})

    def test_adding_a_custom_type_phone_number(self):
        wrapper = TestVCardWrapper()
        wrapper._add_phone_number('custom_type', '0123456789')
        self.assertDictEqual(wrapper.phone_numbers,
                             {'custom_type': ['0123456789']})

    def test_adding_multible_phone_number(self):
        wrapper = TestVCardWrapper()
        wrapper._add_phone_number('work', '0987654321')
        wrapper._add_phone_number('home', '0123456789')
        wrapper._add_phone_number('home', '0112233445')
        self.assertDictEqual(
            wrapper.phone_numbers,
            # The lists are sorted!
            {'home': ['0112233445', '0123456789'], 'work': ['0987654321']})

    def test_adding_preferred_phone_number(self):
        wrapper = TestVCardWrapper()
        wrapper._add_phone_number('home', '0123456789')
        wrapper._add_phone_number('pref,home', '0987654321')
        self.assertDictEqual(
            wrapper.phone_numbers, {'home': ['0123456789'],
                                    'home, pref': ['0987654321']})

    def test_adding_a_simple_email(self):
        wrapper = TestVCardWrapper()
        wrapper.add_email('home', 'foo@bar.net')
        self.assertDictEqual(wrapper.emails, {'home': ['foo@bar.net']})

    def test_adding_a_custom_type_emails(self):
        wrapper = TestVCardWrapper()
        wrapper.add_email('custom_type', 'foo@bar.net')
        self.assertDictEqual(wrapper.emails,
                             {'custom_type': ['foo@bar.net']})

    def test_adding_multible_emails(self):
        wrapper = TestVCardWrapper()
        wrapper.add_email('work', 'foo@bar.net')
        wrapper.add_email('home', 'foo@baz.net')
        wrapper.add_email('home', 'baz@baz.net')
        self.assertDictEqual(
            wrapper.emails,
            # The lists are sorted!
            {'home': ['baz@baz.net', 'foo@baz.net'], 'work': ['foo@bar.net']})

    def test_adding_preferred_emails(self):
        wrapper = TestVCardWrapper()
        wrapper.add_email('home', 'foo@bar.net')
        wrapper.add_email('pref,home', 'foo@baz.net')
        self.assertDictEqual(wrapper.emails, {'home': ['foo@bar.net'],
                                              'home, pref': ['foo@baz.net']})

    def test_adding_a_simple_address(self):
        wrapper = TestVCardWrapper()
        components = ('box', 'extended', 'street', 'code', 'city', 'region',
                      'country')
        wrapper._add_post_address('home', *components)
        expected = {item: item for item in components}
        self.assertDictEqual(wrapper.post_addresses, {'home': [expected]})

    def test_adding_a_custom_type_address(self):
        wrapper = TestVCardWrapper()
        components = ('box', 'extended', 'street', 'code', 'city', 'region',
                      'country')
        wrapper._add_post_address('custom_type', *components)
        expected = {item: item for item in components}
        self.assertDictEqual(wrapper.post_addresses,
                             {'custom_type': [expected]})

    def test_adding_multible_addresses(self):
        wrapper = TestVCardWrapper()
        components = ('box', 'extended', 'street', 'code', 'city', 'region',
                      'country')
        wrapper._add_post_address('work', *['work ' + c for c in components])
        wrapper._add_post_address('home', *['home1 ' + c for c in components])
        wrapper._add_post_address('home', *['home2 ' + c for c in components])
        expected_work = {item: 'work ' + item for item in components}
        expected_home2 = {item: 'home2 ' + item for item in components}
        expected_home1 = {item: 'home1 ' + item for item in components}
        self.assertDictEqual(wrapper.post_addresses,
                             # The lists are sorted!
                             {'home': [expected_home1, expected_home2],
                              'work': [expected_work]})

    def test_adding_preferred_address(self):
        wrapper = TestVCardWrapper()
        components = ('box', 'extended', 'street', 'code', 'city', 'region',
                      'country')
        wrapper._add_post_address('home', *['home1 ' + c for c in components])
        wrapper._add_post_address('pref,home',
                                  *['home2 ' + c for c in components])
        expected_work = {item: 'work ' + item for item in components}
        expected_home2 = {item: 'home2 ' + item for item in components}
        expected_home1 = {item: 'home1 ' + item for item in components}
        self.assertDictEqual(
            wrapper.post_addresses, {'home': [expected_home1],
                                     'home, pref': [expected_home2]})


class OtherProperties(unittest.TestCase):

    def test_setting_and_getting_organisations(self):
        # also test that organisations are returned in sorted order
        wrapper = TestVCardWrapper()
        org1 = ["Org", "Sub1", "Sub2"]
        org2 = ["Org2", "Sub3"]
        org3 = ["Foo", "Bar", "Baz"]
        wrapper._add_organisation(org1)
        wrapper._add_organisation(org2)
        wrapper._add_organisation(org3)
        self.assertListEqual(wrapper.organisations, [org3, org1, org2])

    def test_setting_org_in_different_ways_for_refactoring(self):
        wrapper1 = TestVCardWrapper()
        wrapper2 = TestVCardWrapper()
        wrapper1._add_organisation('foo')
        wrapper2._add_organisation(['foo'])
        self.assertEqual(wrapper1.organisations, wrapper2.organisations)

    def test_setting_and_getting_titles(self):
        wrapper = TestVCardWrapper()
        wrapper._add_title('Foo')
        wrapper._add_title('Bar')
        self.assertListEqual(wrapper.titles, ['Bar', 'Foo'])

    def test_setting_and_getting_roles(self):
        wrapper = TestVCardWrapper()
        wrapper._add_role('Foo')
        wrapper._add_role('Bar')
        self.assertListEqual(wrapper.roles, ['Bar', 'Foo'])

    def test_setting_and_getting_nicks(self):
        wrapper = TestVCardWrapper()
        wrapper._add_nickname('Foo')
        wrapper._add_nickname('Bar')
        self.assertListEqual(wrapper.nicknames, ['Bar', 'Foo'])

    def test_setting_and_getting_notes(self):
        wrapper = TestVCardWrapper()
        wrapper._add_note('First long note')
        wrapper._add_note('Second long note\nwith newline')
        self.assertListEqual(wrapper.notes, ['First long note',
                             'Second long note\nwith newline'])

    def test_setting_and_getting_webpages(self):
        wrapper = TestVCardWrapper()
        wrapper._add_webpage('https://github.com/scheibler/khard')
        wrapper._add_webpage('http://example.com')
        self.assertListEqual(wrapper.webpages, ['http://example.com',
                             'https://github.com/scheibler/khard'])

    def test_setting_and_getting_categories(self):
        wrapper = TestVCardWrapper()
        wrapper._add_category(["rfc", "address book"])
        wrapper._add_category(["coding", "open source"])
        self.assertListEqual(wrapper.categories,
                             [["coding", "open source"],
                              ["rfc", "address book"]])


class ABLabels(unittest.TestCase):

    def test_setting_and_getting_webpage_ablabel(self):
        wrapper = TestVCardWrapper()
        wrapper._add_webpage({'github': 'https://github.com/scheibler/khard'})
        wrapper._add_webpage('http://example.com')
        self.assertListEqual(wrapper.webpages, [
            'http://example.com',
            {'github': 'https://github.com/scheibler/khard'}])

    def test_labels_on_structured_values(self):
        vcard = VCardWrapper(_from_file('test/fixture/vcards/labels.vcf'))
        self.assertListEqual(vcard.organisations, [{'Work': ['Test Inc']}])

    def test_setting_fn_from_labelled_org(self):
        wrapper = TestVCardWrapper()
        wrapper._delete_vcard_object("FN")
        wrapper._add_organisation({'Work': ['Test Inc']})
        self.assertEqual(wrapper.formatted_name, 'Test Inc')



"""Module to make it possible to import the test folder as a python package and
hence make it possible to run all unittests from the top level direcotry with

        python -m unittest [discover]

and

        python setup.py test

"""



"""Tests for the cli module"""

from argparse import ArgumentTypeError
import tempfile
import unittest
from unittest import mock

from khard import cli
from khard import query

from .helpers import mock_stream


class TestFieldsArgument(unittest.TestCase):

    def test_works_when_choices_match(self):
        t = cli.FieldsArgument("a", "b")
        actual = t("a,b")
        expected = ["a", "b"]
        self.assertListEqual(actual, expected)

    def test_raises_exception_when_choices_dont_match(self):
        t = cli.FieldsArgument("a", "b")
        with self.assertRaises(ArgumentTypeError):
            t("a,c")

    def test_case_does_not_matter(self):
        t = cli.FieldsArgument("a", "b")
        actual = t("a,B")
        expected = ["a", "b"]
        self.assertListEqual(actual, expected)

    def test_only_first_component_must_match_choices_with_nested(self):
        t = cli.FieldsArgument("a", "b", nested=True)
        actual = t("a.c,b")
        expected = ["a.c", "b"]
        self.assertListEqual(actual, expected)


@mock.patch.dict('os.environ', KHARD_CONFIG='test/fixture/minimal.conf')
class TestParseArgs(unittest.TestCase):

    foo = query.TermQuery("foo")
    bar = query.TermQuery("bar")
    baz = query.TermQuery("baz")
    uid = query.FieldQuery("uid", "foo")

    def test_normal_search_terms_create_term_queries(self):
        expected = self.foo
        args, _config = cli.parse_args(['list', 'foo'])
        actual = args.search_terms
        self.assertEqual(expected, actual)

    def test_uid_options_create_uid_queries(self):
        expected = self.uid
        args, _config = cli.parse_args(['list', '--uid=foo'])
        actual = args.search_terms
        self.assertEqual(expected, actual)

    def test_multible_search_terms_generate_and_queries(self):
        expected = query.AndQuery(self.foo, self.bar)
        args, _config = cli.parse_args(['list', 'foo', 'bar'])
        actual = args.search_terms
        self.assertEqual(expected, actual)

    def test_no_search_terms_create_an_any_query(self):
        expected = query.AnyQuery()
        args, _config = cli.parse_args(['list'])
        actual = args.search_terms
        self.assertEqual(expected, actual)

    def test_target_search_terms_are_typed(self):
        args, _config = cli.parse_args(['merge', '--target=foo', 'bar'])
        self.assertEqual(self.foo, args.target_contact)
        self.assertEqual(self.bar, args.source_search_terms)

    def test_second_target_search_term_overrides_first(self):
        args, _config = cli.parse_args(['merge', '--target=foo',
                                        '--target=bar', 'baz'])
        self.assertEqual(self.bar, args.target_contact)
        self.assertEqual(self.baz, args.source_search_terms)

    def test_target_uid_option_creates_uid_queries(self):
        args, _config = cli.parse_args(['merge', '--target-uid=foo', 'bar'])
        self.assertEqual(self.uid, args.target_contact)
        self.assertEqual(self.bar, args.source_search_terms)

    def test_uid_option_is_combined_with_search_terms_for_merge_command(self):
        args, _config = cli.parse_args(['merge', '--uid=foo', '--target=bar'])
        self.assertEqual(self.uid, args.source_search_terms)
        self.assertEqual(self.bar, args.target_contact)

    def test_uid_and_free_search_terms_produce_a_conflict(self):
        with self.assertRaises(SystemExit):
            with mock_stream("stderr"):  # just silence stderr
                cli.parse_args(['list', '--uid=foo', 'bar'])

    def test_target_uid_and_free_target_search_terms_produce_a_conflict(self):
        with self.assertRaises(SystemExit):
            with mock_stream("stderr"):  # just silence stderr
                cli.parse_args(['merge', '--target-uid=foo', '--target=bar'])

    def test_no_target_specification_results_in_an_any_query(self):
        expected = query.AnyQuery()
        args, _config = cli.parse_args(['merge'])
        actual = args.target_contact
        self.assertEqual(expected, actual)

    def test_add_email_defaults_to_from_lowercase(self):
        args, _config = cli.parse_args(["add-email"])
        actual = args.headers
        self.assertEqual(["from"], actual)

    def test_add_email_from_field(self):
        args, _config = cli.parse_args(["add-email", "-H", "from"])
        actual = args.headers
        self.assertEqual(["from"], actual)

    def test_add_email_another_field(self):
        args, _config = cli.parse_args(["add-email", "-H", "OtHer"])
        actual = args.headers
        self.assertEqual(["other"], actual)

    def test_add_email_multiple_headers_separate_args_takes_last(self):
        args, _config = cli.parse_args(
            ["add-email", "-H", "OtHer", "-H", "myfield"])
        actual = args.headers
        self.assertEqual(["myfield"], actual)

    def test_add_email_multiple_headers_comma_separated(self):
        args, _config = cli.parse_args(
            ["add-email", "-H", "OtHer,myfield,from"])
        actual = args.headers
        self.assertEqual(["other", "myfield", "from"], actual)

    def test_exit_user_friendly_without_config_file(self):
        with self.assertRaises(SystemExit):
            cli.parse_args(["-c", "/this file should hopefully never exist."])

    def test_exit_user_friendly_without_contacts_folder(self):
        with tempfile.NamedTemporaryFile("w", delete=False) as config:
            config.write("""[general]
                            editor = editor
                            merge_editor = merge_editor
                            [addressbooks]
                            [[tmp]]
                            path = /this file should hopefully never exist.
                            """)
            config.flush()
            with self.assertRaises(SystemExit):
                cli.init(["-c", config.name, "ls"])



"""Tests for the config module."""
# pylint: disable=missing-docstring

import logging
import os.path
import tempfile
import unittest
import unittest.mock as mock

import configobj

from khard import config


class LoadingConfigFile(unittest.TestCase):

    def test_load_non_existing_file_fails(self):
        filename = "I hope this file never exists"
        with self.assertRaises(OSError) as cm:
            config.Config._load_config_file(filename)
        self.assertTrue(str(cm.exception).startswith('Config file not found:'))

    def test_uses_khard_config_environment_variable(self):
        filename = "this is some very random string"
        with mock.patch.dict("os.environ", clear=True, KHARD_CONFIG=filename):
            with mock.patch("configobj.ConfigObj", dict):
                ret = config.Config._load_config_file("")
        self.assertEqual(ret['infile'], filename)

    def test_uses_xdg_config_home_environment_variable(self):
        prefix = "this is some very random string"
        with mock.patch.dict("os.environ", clear=True, XDG_CONFIG_HOME=prefix):
            with mock.patch("configobj.ConfigObj", dict):
                ret = config.Config._load_config_file("")
        expected = os.path.join(prefix, 'khard', 'khard.conf')
        self.assertEqual(ret['infile'], expected)

    def test_uses_config_dir_if_environment_unset(self):
        prefix = "this is some very random string"
        with mock.patch.dict("os.environ", clear=True, HOME=prefix):
            with mock.patch("configobj.ConfigObj", dict):
                ret = config.Config._load_config_file("")
        expected = os.path.join(prefix, '.config', 'khard', 'khard.conf')
        self.assertEqual(ret['infile'], expected)

    def test_load_empty_file_fails(self):
        with tempfile.NamedTemporaryFile() as name:
            with self.assertLogs(level=logging.ERROR):
                with self.assertRaises(config.ConfigError):
                    config.Config(name)

    @mock.patch.dict('os.environ', EDITOR='editor', MERGE_EDITOR='meditor')
    def test_load_minimal_file_by_name(self):
        cfg = config.Config("test/fixture/minimal.conf")
        self.assertEqual(cfg.editor, ["editor"])
        self.assertEqual(cfg.merge_editor, "meditor")


class ConfigPreferredVcardVersion(unittest.TestCase):

    def test_default_value_is_3(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.preferred_vcard_version, "3.0")

    def test_set_preferred_version(self):
        c = config.Config("test/fixture/minimal.conf")
        c.preferred_vcard_version = "11"
        self.assertEqual(c.preferred_vcard_version, "11")


class Defaults(unittest.TestCase):

    def test_debug_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.debug)

    def test_default_action_defaults_to_none(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertIsNone(c.default_action)

    def test_reverse_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.reverse)

    def test_group_by_addressbook_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.group_by_addressbook)

    def test_show_nicknames_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.show_nicknames)

    def test_show_uids_defaults_to_true(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertTrue(c.show_uids)

    def test_show_kinds_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.show_kinds)

    def test_sort_defaults_to_first_name(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.sort, 'first_name')

    def test_display_defaults_to_first_name(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.display, 'first_name')

    def test_localize_dates_defaults_to_true(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertTrue(c.localize_dates)

    def test_preferred_phone_number_type_defaults_to_pref(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertListEqual(c.preferred_phone_number_type, ['pref'])

    def test_preferred_email_address_type_defaults_to_pref(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertListEqual(c.preferred_email_address_type, ['pref'])

    def test_private_objects_defaults_to_empty(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertListEqual(c.private_objects, [])

    def test_search_in_source_files_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.search_in_source_files)

    def test_skip_unparsable_defaults_to_false(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertFalse(c.skip_unparsable)

    def test_preferred_version_defaults_to_3(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.preferred_vcard_version, '3.0')

    @mock.patch.dict('os.environ', clear=True)
    def test_editor_defaults_to_vim(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.editor, ['vim'])

    @mock.patch.dict('os.environ', clear=True)
    def test_merge_editor_defaults_to_vimdiff(self):
        c = config.Config("test/fixture/minimal.conf")
        self.assertEqual(c.merge_editor, 'vimdiff')


class Validation(unittest.TestCase):

    @staticmethod
    def _template(section, key, value):
        configspec = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                  'khard', 'data', 'config.spec')
        c = configobj.ConfigObj(configspec=configspec)
        c['general'] = {}
        c['vcard'] = {}
        c['contact table'] = {}
        c['addressbooks'] = {'test': {'path': '/tmp'}}
        c[section][key] = value
        return c

    def test_rejects_invalid_default_actions(self):
        action = 'this is not a valid action'
        conf = self._template('general', 'default_action', action)
        with self.assertLogs(level=logging.ERROR):
            with self.assertRaises(config.ConfigError):
                config.Config._validate(conf)

    def test_rejects_unparsable_editor_commands(self):
        editor = 'editor --option "unparsable because quotes are missing'
        conf = self._template('general', 'editor', editor)
        with self.assertLogs(level=logging.ERROR):
            with self.assertRaises(config.ConfigError):
                config.Config._validate(conf)

    def test_rejects_private_objects_with_strange_chars(self):
        obj = 'X-VCÄRD-EXTENSIÖN'
        conf = self._template('vcard', 'private_objects', obj)
        with self.assertLogs(level=logging.ERROR):
            with self.assertRaises(config.ConfigError):
                config.Config._validate(conf)

    def test_rejects_private_objects_starting_with_minus(self):
        obj = '-INVALID-'
        conf = self._template('vcard', 'private_objects', obj)
        with self.assertLogs(level=logging.ERROR):
            with self.assertRaises(config.ConfigError):
                config.Config._validate(conf)


if __name__ == "__main__":
    unittest.main()



"""Tests for the action class"""
# pylint: disable=missing-docstring

import unittest

from khard import actions

action = 'list'
alias = 'ls'
unknown = 'this is not an action or an alias'


class Action(unittest.TestCase):

    def test_get_action_resolves_aliases(self):
        self.assertEqual(action, actions.Actions.get_action(alias))

    def test_get_action_returns_none_for_actions(self):
        self.assertIsNone(actions.Actions.get_action(action))

    def test_get_action_returns_none_for_unknown(self):
        self.assertIsNone(actions.Actions.get_action(unknown))

    def test_get_aliases_reverse_resolves_aliases(self):
        self.assertEqual([alias], actions.Actions.get_aliases(action))

    def test_get_aliases_throws_keyerror_for_aliases(self):
        with self.assertRaises(KeyError):
            actions.Actions.get_aliases(alias)

    def test_get_aliases_throws_keyerror_for_unknown(self):
        with self.assertRaises(KeyError):
            actions.Actions.get_aliases(unknown)

    def test_get_actions_returns_actions(self):
        self.assertIn(action, actions.Actions.get_actions())

    def test_get_actions_does_not_return_aliases(self):
        self.assertNotIn(alias, actions.Actions.get_actions())

    def test_get_all_returns_actions(self):
        self.assertIn(action, actions.Actions.get_all())

    def test_get_all_returns_aliases(self):
        self.assertIn(alias, actions.Actions.get_all())



import unittest

from khard.query import AndQuery, AnyQuery, FieldQuery, NameQuery, NullQuery, \
    OrQuery, TermQuery, parse

from .helpers import TestCarddavObject, load_contact


class TestTermQuery(unittest.TestCase):

    def test_match_if_query_is_anywhere_in_string(self):
        q = TermQuery('bar')
        self.assertTrue(q.match('foo bar baz'))

    def test_query_terms_are_case_insensitive(self):
        q = TermQuery('BAR')
        self.assertTrue(q.match('foo bar baz'))

    def test_match_arguments_are_case_insensitive(self):
        q = TermQuery('bar')
        self.assertTrue(q.match('FOO BAR BAZ'))

    def test_spaces_in_search_subject_are_not_stripped(self):
        q = TermQuery('oob')
        self.assertFalse(q.match('foo bar baz'))

    def test_spaces_in_query_are_not_stripped(self):
        q = TermQuery('foo bar')
        self.assertFalse(q.match('foobar'))


class TestAndQuery(unittest.TestCase):

    def test_matches_if_all_subterms_match(self):
        q1 = TermQuery("a")
        q2 = TermQuery("b")
        q = AndQuery(q1, q2)
        self.assertTrue(q.match("ab"))

    def test_failes_if_at_least_one_subterm_fails(self):
        q1 = TermQuery("a")
        q2 = TermQuery("b")
        q = AndQuery(q1, q2)
        self.assertFalse(q.match("ac"))

    def test_order_does_not_matter(self):
        q1 = TermQuery("a")
        q2 = TermQuery("b")
        q = AndQuery(q1, q2)
        self.assertTrue(q.match("ab"))
        self.assertTrue(q.match("ba"))

class TestOrQuery(unittest.TestCase):

    def test_matches_if_at_least_one_subterm_matchs(self):
        q1 = TermQuery("a")
        q2 = TermQuery("b")
        q = OrQuery(q1, q2)
        self.assertTrue(q.match("ac"))

    def test_failes_if_all_subterms_fail(self):
        q1 = TermQuery("a")
        q2 = TermQuery("b")
        q = OrQuery(q1, q2)
        self.assertFalse(q.match("cd"))

    def test_order_does_not_matter(self):
        q1 = TermQuery("a")
        q2 = TermQuery("b")
        q = OrQuery(q1, q2)
        self.assertTrue(q.match("ab"))
        self.assertTrue(q.match("ba"))


class TestEquality(unittest.TestCase):

    def test_any_queries_are_equal(self):
        self.assertEqual(AnyQuery(), AnyQuery())

    def test_null_queries_are_equal(self):
        self.assertEqual(NullQuery(), NullQuery())

    def test_or_queries_match_after_sorting(self):
        null = NullQuery()
        any = AnyQuery()
        term = TermQuery("foo")
        field = FieldQuery("x", "y")
        first = OrQuery(null, any, term, field)
        second = OrQuery(any, null, field, term)
        self.assertEqual(first, second)

    def test_and_queries_match_after_sorting(self):
        null = NullQuery()
        any = AnyQuery()
        term = TermQuery("foo")
        field = FieldQuery("x", "y")
        first = AndQuery(null, any, term, field)
        second = AndQuery(any, null, field, term)
        self.assertEqual(first, second)


class TestFieldQuery(unittest.TestCase):

    @unittest.expectedFailure
    def test_empty_field_values_match_if_the_field_is_present(self):
        # This test currently fails because the CarddavObject class has all
        # attributes set because they are properties.  So the test in the query
        # class if an attribute is present never fails.
        uid = 'Some Test Uid'
        vcard1 = TestCarddavObject(uid=uid)
        vcard2 = TestCarddavObject()
        query = FieldQuery('uid', '')
        self.assertTrue(query.match(vcard1))
        self.assertFalse(query.match(vcard2))

    def test_empty_field_values_fails_if_the_field_is_absent(self):
        vcard = TestCarddavObject()
        query = FieldQuery('emails', '')
        self.assertFalse(query.match(vcard))

    def test_values_can_match_exact(self):
        uid = 'Some Test Uid'
        vcard = TestCarddavObject(uid=uid)
        query = FieldQuery('uid', uid)
        self.assertTrue(query.match(vcard))

    def test_values_can_match_substrings(self):
        uid = 'Some Test Uid'
        vcard = TestCarddavObject(uid=uid)
        query = FieldQuery('uid', 'e Test U')
        self.assertTrue(query.match(vcard))

    def test_valuess_can_match_case_insensitive(self):
        uid = 'Some Test Uid'
        vcard = TestCarddavObject(uid=uid)
        query1 = FieldQuery('uid', uid.upper())
        query2 = FieldQuery('uid', uid.lower())
        self.assertTrue(query1.match(vcard))
        self.assertTrue(query2.match(vcard))

    def test_match_formatted_name(self):
        vcard = TestCarddavObject(fn='foo bar')
        query = FieldQuery('formatted_name', 'foo')
        self.assertTrue(query.match(vcard))

    def test_match_email(self):
        vcard = load_contact("contact1.vcf")
        query = FieldQuery('emails', 'user@example.com')
        self.assertTrue(query.match(vcard))

    def test_match_birthday(self):
        vcard = load_contact("contact1.vcf")
        query = FieldQuery('birthday', '2018-01-20')
        self.assertTrue(query.match(vcard))

    def test_fail_match_in_other_field(self):
        vcard = load_contact("contact1.vcf")
        query = FieldQuery('formatted_name', 'user@example.com')
        self.assertFalse(query.match(vcard))

    def test_match_email_type(self):
        vcard = load_contact("contact1.vcf")
        query = FieldQuery('emails', 'home')
        self.assertTrue(query.match(vcard))


class TestNameQuery(unittest.TestCase):

    def test_matches_formatted_name_field(self):
        vcard = load_contact("minimal.vcf")
        query = NameQuery("minimal")
        self.assertTrue(query.match(vcard))

    def test_matches_name_field(self):
        vcard = load_contact("nickname.vcf")
        query = NameQuery("smith")
        self.assertTrue(query.match(vcard))

    def test_matches_nickname_field(self):
        vcard = load_contact("nickname.vcf")
        query = NameQuery("mike")
        self.assertTrue(query.match(vcard))

    def test_does_not_match_uid_field(self):
        vcard = load_contact("contact1.vcf")
        query = NameQuery("testuid1")
        self.assertFalse(query.match(vcard))


class TestParser(unittest.TestCase):

    def test_parsing_simple_terms(self):
        string = "foo bar"
        expected = TermQuery(string)
        actual = parse(string)
        self.assertEqual(actual, expected)

    def test_parsing_simple_field_queries(self):
        actual = parse("formatted_name:foo bar")
        expected = FieldQuery("formatted_name", "foo bar")
        self.assertEqual(actual, expected)

    def test_bad_field_name_returns_term_query(self):
        string = "foo:bar"
        actual = parse(string)
        expected = TermQuery(string)
        self.assertEqual(actual, expected)

    def test_field_value_can_be_empty(self):
        actual = parse("formatted_name:")
        expected = FieldQuery("formatted_name", "")
        self.assertEqual(actual, expected)

    def test_field_value_can_contain_colons(self):
        actual = parse("formatted_name:foo:bar")
        expected = FieldQuery("formatted_name", "foo:bar")
        self.assertEqual(actual, expected)

    def test_special_field_name_creates_name_queries(self):
        actual = parse("name:foo")
        expected = NameQuery("foo")
        self.assertEqual(actual, expected)



"""Tests for the address book classes."""
# pylint: disable=missing-docstring

import os
import unittest
from unittest import mock

from khard import address_book, query

from .helpers import TmpAbook


class _AddressBook(address_book.AddressBook):
    """Class for testing the abstract AddressBook base class."""

    def load(self, query=None):
        pass


class AbstractAddressBookSearch(unittest.TestCase):
    """Tests for khard.address_book.AddressBook.search()"""

    def test_search_will_trigger_load_if_not_loaded(self):
        abook = _AddressBook('test')
        load_mock = mock.Mock()
        abook.load = load_mock
        list(abook.search(query.AnyQuery()))
        load_mock.assert_called_once()

    def test_search_will_not_trigger_load_if_loaded(self):
        abook = _AddressBook('test')
        load_mock = mock.Mock()
        abook.load = load_mock
        abook._loaded = True
        list(abook.search(query.AnyQuery()))
        load_mock.assert_not_called()

    def test_search_passes_query_to_load(self):
        abook = _AddressBook('test')
        self.assertFalse(abook._loaded)
        load_mock = mock.Mock()
        abook.load = load_mock
        list(abook.search(query.AnyQuery()))
        load_mock.assert_called_once_with(query.AnyQuery())


class AddressBookCompareUids(unittest.TestCase):

    def test_different_strings(self):
        uid1 = 'abc'
        uid2 = 'xyz'
        expected = 0
        actual = address_book.AddressBook._compare_uids(uid1, uid2)
        self.assertEqual(actual, expected)

    def test_two_simple_strings(self):
        uid1 = 'abcdef'
        uid2 = 'abcxyz'
        expected = 3
        actual = address_book.AddressBook._compare_uids(uid1, uid2)
        self.assertEqual(actual, expected)

    def test_no_error_on_equal_strings(self):
        uid = 'abcdefghij'
        expected = len(uid)
        actual = address_book.AddressBook._compare_uids(uid, uid)
        self.assertEqual(actual, expected)


class VcardAddressBookLoad(unittest.TestCase):

    def test_vcards_without_uid_generate_a_warning(self):
        abook = address_book.VdirAddressBook('test',
                                             'test/fixture/minimal.abook')
        with self.assertLogs(level='WARNING') as cm:
            abook.load()
            messages = ['WARNING:khard.address_book:Card minimal contact from '
                        'address book test has no UID and will not be '
                        'available.']
        self.assertListEqual(cm.output, messages)

    def test_loading_vcards_from_disk(self):
        abook = address_book.VdirAddressBook('test', 'test/fixture/test.abook')
        # At this point we do not really care about the type of abook.contacts,
        # it could be a list or dict or set or whatever.
        self.assertEqual(len(abook.contacts), 0)
        abook.load()
        self.assertEqual(len(abook.contacts), 3)

    def test_search_in_source_files_only_loads_matching_cards(self):
        abook = address_book.VdirAddressBook('test', 'test/fixture/test.abook')
        abook.load(query=query.TermQuery('second'), search_in_source_files=True)
        self.assertEqual(len(abook.contacts), 1)

    def test_loading_unparsable_vcard_fails(self):
        abook = address_book.VdirAddressBook('test',
                                             'test/fixture/broken.abook')
        with self.assertRaises(address_book.AddressBookParseError):
            with self.assertLogs(level='ERROR'):
                abook.load()

    def test_unparsable_files_can_be_skipped(self):
        abook = address_book.VdirAddressBook(
            'test', 'test/fixture/broken.abook', skip=True)
        with self.assertLogs(level='WARNING') as cm:
            abook.load()
        self.assertEqual(cm.output[0],
            'WARNING:khard.carddav_object:Filtering some problematic tags '
            'from test/fixture/broken.abook/unparsable.vcf')
        # FIXME Remove this regex assert when either
        # https://github.com/eventable/vobject/issues/156 is closed or we drop
        # support for python 3.6
        self.assertRegex(cm.output[1],
            'ERROR:khard.address_book:Error: Could not parse file '
            'test/fixture/broken.abook/unparsable.vcf\n'
            'At line [35]: Component VCARD was never closed')
        self.assertEqual(cm.output[2],
            'WARNING:khard.address_book:1 of 1 vCard files of address book '
            'test could not be parsed.')

    @mock.patch.dict("os.environ", clear=True)
    def test_do_not_expand_env_var_that_is_unset(self):
        # Unset env vars shouldn't expand.
        with self.assertRaises(FileNotFoundError):
            address_book.VdirAddressBook(
                "test", "test/fixture/test.abook${}".format("KHARD_FOO"))

    @mock.patch.dict("os.environ", KHARD_FOO="")
    def test_expand_env_var_that_is_empty(self):
        # Env vars set to empty string should expand to empty string.
        abook = address_book.VdirAddressBook(
            "test", "test/fixture/test.abook${}".format("KHARD_FOO"))
        self.assertEqual(abook.path, "test/fixture/test.abook")

    @mock.patch.dict("os.environ", KHARD_FOO="test/fixture")
    def test_expand_env_var_that_is_nonempty(self):
        # Env vars set to nonempty string should expand appropriately.
        abook = address_book.VdirAddressBook(
            "test", "${}/test.abook".format("KHARD_FOO"))
        self.assertEqual(abook.path, "test/fixture/test.abook")


class VcardAddressBookSearch(unittest.TestCase):

    @staticmethod
    def _search(query):
        with TmpAbook(["contact1.vcf", "contact2.vcf"]) as abook:
            return list(abook.search(query))

    def test_uid_query(self):
        q = query.FieldQuery("uid", "testuid1")
        l = self._search(q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, 'testuid1')

    def test_term_query(self):
        q = query.TermQuery("testuid1")
        l = self._search(q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, 'testuid1')

    def test_term_query_matching(self):
        q = query.TermQuery("second contact")
        l = self._search(q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, 'testuid1')

    def test_term_query_failing(self):
        q = query.TermQuery("this does not match")
        l = self._search(q)
        self.assertEqual(len(l), 0)

    def test_copied_from_merge_test_1(self):
        q = query.TermQuery("second")
        l = self._search(q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, 'testuid1')

    def test_copied_from_merge_test_2(self):
        q = query.TermQuery("third")
        l = self._search(q)
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0].uid, 'testuid2')


class AddressBookGetShortUidDict(unittest.TestCase):

    def test_uniqe_uid_also_reslts_in_shortend_uid_in_short_uid_dict(self):
        contacts = {'uid123': None}
        abook = _AddressBook('test')
        abook.contacts = contacts
        abook._loaded = True
        short_uids = abook.get_short_uid_dict()
        self.assertEqual(len(short_uids), 1)
        short_uid, contact = short_uids.popitem()
        self.assertEqual(short_uid, 'u')


class ReportedBugs(unittest.TestCase):

    def test_issue_159_uid_search_doesnt_return_items_twice(self):
        # This was the first half of bug report #159.
        abook = address_book.VdirAddressBook('test', 'test/fixture/test.abook')
        c = abook.search(query.TermQuery('testuid1'))
        self.assertEqual(len(list(c)), 1)



"""Loading and validation of the configuration file"""

from argparse import Namespace
import locale
import logging
import os
import re
import shlex
from typing import Iterable, Dict, List, Optional, Union

import configobj
try:
    # since configobj 5.1
    from configobj import validate
except ImportError:
    import validate

from .actions import Actions
from .address_book import AddressBookCollection, AddressBookNameError, \
    VdirAddressBook
from .query import Query


logger = logging.getLogger(__name__)


class ConfigError(Exception):
    """Errors during config file parsing"""


def validate_command(value: List[str]) -> List[str]:
    """Special validator to check shell commands

    The input must either be a list of strings or a string that shlex.split can
    parse into such.

    :param value: the config value to validate
    :returns: the command after validation
    :raises: validate.ValidateError
    """
    logger.debug("validating %s", value)
    try:
        return validate.is_string_list(value)
    except validate.VdtTypeError:
        logger.debug('continue with %s', value)
        if isinstance(value, str):
            try:
                return shlex.split(value)
            except ValueError as err:
                raise validate.ValidateError(
                    'Error when parsing shell command "{}": {}'.format(
                        value, err))
        raise


def validate_action(value: str) -> str:
    """Check that the given value is a valid action.

    :param value: the config value to check
    :returns: the same value
    :raises: validate.ValidateError
    """
    return validate.is_option(value, *Actions.get_actions())


def validate_private_objects(value: List[str]) -> List[str]:
    """Check that the private objects are reasonable

    :param value: the config value to check
    :returns: the list of private objects
    :raises: validate.ValidateError
    """
    value = validate.is_string_list(value)
    for obj in value:
        if re.search("[^a-z0-9-]", obj, re.IGNORECASE):
            raise validate.ValidateError(
                'Private objects may only contain letters, digits and the'
                ' \"-\" character.')
        if obj.startswith("-") or obj.endswith("-"):
            raise validate.ValidateError(
                "A \"-\" in a private object label must be at least "
                "surrounded by one letter or digit.")
    return value


class Config:
    """Parse and validate the config file with configobj."""

    supported_vcard_versions = ("3.0", "4.0")

    def __init__(self, config_file: Optional[str] = None) -> None:
        self.config: configobj.ConfigObj
        self.abooks: AddressBookCollection
        locale.setlocale(locale.LC_ALL, '')
        config = self._load_config_file(config_file)
        self.config = self._validate(config)
        self._set_attributes()

    @classmethod
    def _load_config_file(cls, config_file: Optional[str]
                          ) -> configobj.ConfigObj:
        """Find and load the config file.

        :param config_file: the path to the config file to load
        :returns: the loaded config file
        """
        if not config_file:
            xdg_config_home = os.getenv("XDG_CONFIG_HOME",
                                        os.path.expanduser("~/.config"))
            config_file = os.getenv("KHARD_CONFIG", os.path.join(
                xdg_config_home, "khard", "khard.conf"))
        configspec = os.path.join(os.path.dirname(__file__),
                                  'data', 'config.spec')
        try:
            return configobj.ConfigObj(
                infile=config_file, configspec=configspec,
                interpolation=False, file_error=True)
        except configobj.ConfigObjError as err:
            raise ConfigError(str(err))

    @staticmethod
    def _validate(config: configobj.ConfigObj) -> configobj.ConfigObj:
        vdr = validate.Validator()
        vdr.functions.update({'command': validate_command,
                              'action': validate_action,
                              'private_objects': validate_private_objects})
        result = config.validate(vdr, preserve_errors=True)
        result = configobj.flatten_errors(config, result)
        if not config['addressbooks'].keys():
            result.append((['addressbooks'], '__any__',
                           'No address book entries available'))
        for path, key, exception in result:
            logger.error("Error in config file, %s: %s",
                         ".".join([*path, key]), exception)
        if result:
            raise ConfigError
        return config

    def _set_attributes(self) -> None:
        """Set the attributes from the internal config instance on self."""
        general = self.config["general"]
        self.debug = general["debug"]
        self.editor = (
            general["editor"] or shlex.split(os.environ.get("EDITOR", "vim"))
        )
        self.merge_editor = general["merge_editor"] \
            or os.environ.get("MERGE_EDITOR", "vimdiff")
        self.default_action = general["default_action"]
        table = self.config["contact table"]
        vcard = self.config["vcard"]
        self.sort = table["sort"]
        # if display by name attribute is not present in the config file use
        # the sort attribute value for backwards compatibility
        self.display = table.get("display", self.sort)
        self.localize_dates = table['localize_dates']
        self.private_objects = vcard['private_objects']
        self.preferred_vcard_version = vcard['preferred_version']
        self.search_in_source_files = vcard['search_in_source_files']
        self.skip_unparsable = vcard['skip_unparsable']
        self.group_by_addressbook = table['group_by_addressbook']
        self.reverse = table['reverse']
        self.show_nicknames = table['show_nicknames']
        self.preferred_email_address_type = table['preferred_email_address_type']
        self.preferred_phone_number_type = table['preferred_phone_number_type']
        self.show_uids = table['show_uids']
        self.show_kinds = table['show_kinds']

    def init_address_books(self) -> None:
        """Initialize the internal address book collection.

        This method should only be called *after* merging in the command line
        options as they can hold some options that are relevant for the loading
        of the address books.
        """
        section = self.config['addressbooks']
        kwargs = {'private_objects': self.private_objects,
                  'localize_dates': self.localize_dates,
                  'skip': self.skip_unparsable}
        try:
            self.abooks = AddressBookCollection(
                "tmp", [VdirAddressBook(name, section[name]['path'], **kwargs)
                        for name in section])
        except OSError as err:
            raise ConfigError(str(err))

    def get_address_books(self, names: Iterable[str], queries: Dict[str, Query]
                          ) -> AddressBookCollection:
        """Load all address books with the given names.

        :param names: the address books to load
        :param queries: a mapping of address book names to search queries
        :returns: the loaded address books
        """
        all_names = {str(book) for book in self.abooks}
        if not names:
            names = all_names
        elif not all_names.issuperset(names):
            raise AddressBookNameError(
                "The following address books are not defined: {}".format(
                    ', '.join(set(names) - all_names)))
        # load address books which are defined in the configuration file
        collection = AddressBookCollection("tmp", [self.abooks[name]
                                                   for name in names])
        # We can not use AddressBookCollection.load here because we want to
        # select the collection based on the address book.
        for abook in collection:
            abook.load(queries[abook.name], self.search_in_source_files)
        return collection

    def merge(self, other: Union[configobj.ConfigObj, Dict]) -> None:
        """Merge the config with some other dict or ConfigObj

        :param other: the other dict or ConfigObj to merge into self
        :returns: None
        """
        self.config.merge(other)
        self._validate(self.config)
        self._set_attributes()

    def merge_args(self, args: Namespace) -> None:
        """Merge options from a flat argparse object.

        :param argparse.Namespace args: the parsed arguments to incorperate
        """
        skel = {'general': ['debug'],
                'contact table': ['reverse', 'group_by_addressbook',
                                  'display', 'sort'],
                'vcard': ['search_in_source_files', 'skip_unparsable',
                          'preferred_version'],
                }
        merge = {sec: {key: getattr(args, key) for key in opts
                       if key in args and getattr(args, key) is not None}
                 for sec, opts in skel.items()}
        logger.debug('Merging in %s', merge)
        self.merge(merge)
        logger.debug('Merged: %s', vars(self))



"""Tests for editing files and contacts in an external editor"""

from contextlib import contextmanager
import datetime
import unittest
from unittest import mock

from khard.helpers.interactive import Editor, EditState

from .helpers import mock_stream


class EditFiles(unittest.TestCase):

    t1 = datetime.datetime(2021, 1, 1, 12, 21, 42)
    t2 = datetime.datetime(2021, 1, 1, 12, 21, 43)
    editor = Editor("edit", "merge")

    @staticmethod
    @contextmanager
    def _mock_popen(returncode=0):
        """Mock the subprocess.Popen class, set the returncode attribute of the
        child process object."""
        child_process = mock.Mock()
        child_process.returncode = returncode
        Popen = mock.Mock(return_value=child_process)
        with mock.patch("subprocess.Popen", Popen) as popen:
            yield popen

    @staticmethod
    def _edit_files(write="changed"):
        """Mock function for khar.helpers.interactive.Editor.edit_files

        Create a function that will write the specified text to all files
        passed as arguments.
        """
        def edit_files(self, *files):
            for f in files:
                with open(f, "w") as fp:
                    fp.write(write)
        return edit_files

    def test_calls_subprocess_popen_with_editor_for_one_args(self):
        with self._mock_popen() as popen:
            with mock.patch("khard.helpers.interactive.Editor._mtime",
                            mock.Mock(return_value=self.t1)):
                self.editor.edit_files("file")
        popen.assert_called_with(["edit", "file"])

    def test_calls_subprocess_popen_with_merge_editor_for_two_args(self):
        with self._mock_popen() as popen:
            with mock.patch("khard.helpers.interactive.Editor._mtime",
                            mock.Mock(return_value=self.t1)):
                self.editor.edit_files("file1", "file2")
        popen.assert_called_with(["merge", "file1", "file2"])

    def test_failing_external_command_returns_aborted_state(self):
        with self._mock_popen(1):
            with mock.patch("khard.helpers.interactive.Editor._mtime",
                            mock.Mock(return_value=self.t1)):
                actual = self.editor.edit_files("file")
        self.assertEqual(actual, EditState.aborted)

    def test_returns_state_modiefied_if_timestamp_does_change(self):
        with self._mock_popen():
            with mock.patch("khard.helpers.interactive.Editor._mtime",
                            mock.Mock(side_effect=[self.t1, self.t2])):
                actual = self.editor.edit_files("file")
        self.assertEqual(actual, EditState.modified)

    def test_returns_state_unmodiefied_if_timestamp_does_not_change(self):
        with self._mock_popen():
            with mock.patch("khard.helpers.interactive.Editor._mtime",
                            mock.Mock(side_effect=[self.t1, self.t1])):
                actual = self.editor.edit_files("file")
        self.assertEqual(actual, EditState.unmodified)

    def test_editing_templates(self):
        t1 = "some: yaml\ndocument: true\n"
        with mock.patch("khard.helpers.interactive.Editor.edit_files",
                        self._edit_files()):
            actual = self.editor.edit_templates(lambda x: x, t1)
        self.assertEqual(actual, "changed")

    def test_exception_from_yaml_conversion_is_caught(self):
        t1 = "key: value\n"
        with mock.patch("khard.helpers.interactive.Editor.edit_files",
                        self._edit_files()):
            with mock.patch("khard.helpers.interactive.confirm",
                            mock.Mock(return_value=False)) as confirm:
                with mock_stream():  # hide stdout in test
                    actual = self.editor.edit_templates(
                        mock.Mock(side_effect=ValueError), t1)
        self.assertIsNone(actual)
        confirm.assert_called_once()



"""Tests for the CarddavObject class from the carddav module."""
# pylint: disable=missing-docstring

import base64
import datetime
import unittest
from unittest import mock

from khard.carddav_object import CarddavObject


class CarddavObjectFormatDateObject(unittest.TestCase):

    def test_format_date_object_will_not_touch_strings(self):
        expected = 'untouched string'
        actual = CarddavObject._format_date_object(expected, False)
        self.assertEqual(actual, expected)

    def test_format_date_object_with_simple_date_object(self):
        d = datetime.datetime(2018, 2, 13)
        actual = CarddavObject._format_date_object(d, False)
        self.assertEqual(actual, '2018-02-13')

    def test_format_date_object_with_simple_datetime_object(self):
        d = datetime.datetime(2018, 2, 13, 0, 38, 31)
        with mock.patch('time.timezone', -7200):
            actual = CarddavObject._format_date_object(d, False)
        self.assertEqual(actual, '2018-02-13T00:38:31+02:00')

    def test_format_date_object_with_date_1900(self):
        d = datetime.datetime(1900, 2, 13)
        actual = CarddavObject._format_date_object(d, False)
        self.assertEqual(actual, '--02-13')


class AltIds(unittest.TestCase):

    def test_altids_are_read(self):
        card = CarddavObject.from_file(None, 'test/fixture/vcards/altid.vcf')
        expected = 'one representation'
        self.assertEqual(expected, card.get_first_name_last_name())


class Photo(unittest.TestCase):

    """Tests related to the PHOTO property of vCards"""

    PNG_HEADER = b'\x89PNG\r\n\x1a\n'

    def test_parsing_base64_ecoded_photo_vcard_v3(self):
        c = CarddavObject.from_file(None, 'test/fixture/vcards/photov3.vcf')
        self.assertEqual(c.vcard.photo.value[:8], self.PNG_HEADER)

    def test_parsing_base64_ecoded_photo_vcard_v4(self):
        c = CarddavObject.from_file(None, 'test/fixture/vcards/photov4.vcf')
        uri_stuff, data = c.vcard.photo.value.split(',')
        self.assertEqual(uri_stuff, 'data:image/png;base64')
        data = base64.decodebytes(data.encode())
        self.assertEqual(data[:8], self.PNG_HEADER)



# -*- coding: utf-8 -*-

# tutorials:
#  - https://packaging.python.org/en/latest/distributing.html
#  - https://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty/
#  - https://gehrcke.de/2014/02/distributing-a-python-command-line-application/

from setuptools import setup

with open('README.md', 'rb') as f:
    readme = f.read().decode("utf-8")

setup(
    name='khard',
    author='Eric Scheibler',
    author_email='email@eric-scheibler.de',
    url='https://github.com/lucc/khard/',
    description='A console address book manager',
    long_description=readme,
    long_description_content_type='text/markdown',
    license='GPL',
    keywords='vcard console addressbook',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Topic :: Utilities",
        "Topic :: Communications :: Email :: Address Book",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Intended Audience :: End Users/Desktop",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3 :: Only",
    ],
    install_requires=[
        'atomicwrites',
        'configobj',
        'ruamel.yaml',
        'unidecode',
        'vobject'
    ],
    extras_require={'doc': [
        'sphinx',
        'sphinx-autoapi',
        'sphinx-autodoc-typehints'
    ]},
    use_scm_version={'write_to': 'khard/version.py'},
    setup_requires=['setuptools_scm'],
    packages=['khard', 'khard.helpers'],
    package_data={'khard': ['data/*']},
    entry_points={'console_scripts': ['khard = khard.khard:main']},
    test_suite="test",
    # we use type annotations of unset variables which needs 3.6
    python_requires=">=3.7",
    include_package_data=True,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# khard documentation build configuration file, created by
# sphinx-quickstart on Sun Jan 14 10:35:27 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import pathlib
import sys
sys.path.insert(0, os.path.abspath('../..'))
from khard.version import version as original_version


# update the template file for the docs if necessary
def update_template_file():
    here = pathlib.Path(__file__).parent
    src = here.parent.parent/'khard'/'data'/'template.yaml'
    dest = here/'examples'/'template.yaml'
    if not dest.exists() or src.stat().st_ctime > dest.stat().st_ctime:
        dest.write_text(src.read_text().format(
            "\n    Jabber  : \n    Skype   : \n    Twitter : "))


update_template_file()
del update_template_file


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'autoapi.extension',  # https://sphinx-autoapi.readthedocs.io/en/latest/
    'sphinx.ext.autodoc',  # https://pypi.org/project/sphinx-autodoc-typehints/
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.todo',
    'sphinx_autodoc_typehints',  # https://pypi.org/project/sphinx-autodoc-typehints/
]

autoapi_type = 'python'
autoapi_dirs = ['../../khard']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'khard'
copyright = '2020, Eric Scheibler'
author = 'Eric Scheibler'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '.'.join(original_version.split('.')[0:2])
# The full version, including alpha/beta/rc tags.
release = original_version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': [
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'kharddoc'


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('man/khard', 'khard', 'Console address book manager', '', 1),
    ('man/khard.conf', 'khard.conf', 'configuration file for khard', '', 5),
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'khard', 'khard Documentation',
     author, 'khard', 'One line description of project.',
     'Miscellaneous'),
]



"""Some helper functions for khard"""

from datetime import datetime
import pathlib
import random
import string
from typing import Any, Dict, List, Optional, Sequence, Union

from ruamel.yaml.scalarstring import LiteralScalarString
from .typing import list_to_string


def pretty_print(table: List[List[str]], justify: str = "L") -> str:
    """Converts a list of lists into a string formatted like a table
    with spaces separating fields and newlines separating rows"""
    # support for multiline columns
    line_break_table = []
    for row in table:
        # get line break count
        most_line_breaks_in_row = 0
        for col in row:
            if str(col).count("\n") > most_line_breaks_in_row:
                most_line_breaks_in_row = col.count("\n")
        # fill table rows
        for index in range(0, most_line_breaks_in_row+1):
            line_break_row = []
            for col in row:
                try:
                    line_break_row.append(str(col).split("\n")[index])
                except IndexError:
                    line_break_row.append("")
            line_break_table.append(line_break_row)
    # replace table variable
    table = line_break_table
    # get width for every column
    column_widths = [0] * len(table[0])
    offset = 3
    for row in table:
        for index, col in enumerate(row):
            width = len(str(col))
            if width > column_widths[index]:
                column_widths[index] = width
    table_row_list = []
    for row in table:
        single_row_list = []
        for col_index, col in enumerate(row):
            if justify == "R":  # justify right
                formated_column = str(col).rjust(column_widths[col_index] +
                                                 offset)
            elif justify == "L":  # justify left
                formated_column = str(col).ljust(column_widths[col_index] +
                                                 offset)
            elif justify == "C":  # justify center
                formated_column = str(col).center(column_widths[col_index] +
                                                  offset)
            single_row_list.append(formated_column)
        table_row_list.append(' '.join(single_row_list))
    return '\n'.join(table_row_list)


def get_random_uid() -> str:
    return ''.join([random.choice(string.ascii_lowercase + string.digits)
                    for _ in range(36)])

def yaml_clean(value: Union[str, Sequence, Dict[str, Any], None]
              ) -> Union[Sequence, str, Dict[str, Any], LiteralScalarString,
                         None]:
    """
    sanitize yaml values according to some simple principles:
      1. empty values are none, so ruamel does not print an empty list/str
      2. list with only one item become this item
      3. multiline strings use the YAML literal style:
         https://yaml.org/spec/1.2.2/#literal-style

    :param value: the value to be sanitized
    :returns: the sanitized value
    """
    # special case for empty values
    if not value:
        return None

    if isinstance(value, list):
        # special case for single item lists:
        if len(value) == 1 and isinstance(value[0], str):
            return value[0]
    elif isinstance(value, str):
        if "\n" in value:
            return LiteralScalarString(value)

    return value


def yaml_dicts(
        data: Optional[Dict[str, Any]],
        defaults: Union[Dict[str, Any], List[str], None] = None
    ) -> Optional[Dict[str, Any]]:
    """
    format a dict according to template, if empty use specified defaults

    :param data: dict of contact data with keys as types and values as values.
    :param defaults: default dict to be used if data is empty.
    :returns: dict of types and values.
    """
    if not data:
        if isinstance(defaults, list):
            return {key: None for key in defaults}

        return defaults

    data_dict = {}
    for key, value_list in sorted(data.items(), key=lambda k: k[0].lower()):
        data_dict[key] = value_list[0] if len(value_list) == 1 else value_list

    return data_dict


def yaml_addresses(addresses: Optional[Dict[str, Any]],
                   address_properties: List[str],
                   defaults: Optional[List[str]] = None
                  ) -> Optional[Dict[str, Any]]:
    """
    build a dict from an address, using a list of properties, an address has.

    :param addresses: dict of addresses with existing properties
    :param address_properties: list of properties that make up an address
    :param defaults: list of address types
    :returns: dict of address types with all properties
    """
    if not addresses:
        if not defaults:
            return None

        address_fields = {key: None for key in address_properties}
        return {address_type: address_fields for address_type in defaults}

    address_dict = {}
    for address_type, address in addresses.items():
        if isinstance(address, list):
            address = address[0]
        address_dict[address_type] = {
            key: yaml_clean(address.get(f"{key[0].lower()}{key[1:]}"))
            for key in address_properties
        }
    return address_dict


def yaml_anniversary(anniversary: Union[str, datetime, None],
                     version: str) -> Optional[str]:
    """
    format an anniversary according to its contents and the VCard version.

    :param anniversary: a string or a datetime object, that is the anniversary
    :param version: the VCard version to format for
    :returns: the formatted date string
    """
    if not anniversary:
        return None

    if isinstance(anniversary, datetime):
        if (version == "4.0" and anniversary.year == 1900
                             and anniversary.month != 0
                             and anniversary.day != 0
                             and anniversary.hour == 0
                             and anniversary.minute == 0
                             and anniversary.second == 0):
            return anniversary.strftime("--%m-%d")

        time_zone = anniversary.tzname()
        if ((time_zone and time_zone[3:])
                or anniversary.hour != 0
                or anniversary.minute != 0
                or anniversary.second != 0):
            return anniversary.isoformat()

        return anniversary.strftime("%F")

    # default case: anniversary is a string
    return anniversary


def convert_to_yaml(name: str, value: Union[None, str, List], indentation: int,
                    index_of_colon: int, show_multi_line_character: bool
                    ) -> List[str]:
    """converts a value list into yaml syntax

    :param name: name of object (example: phone)
    :param value: object contents
    :type value: str, list(str), list(list(str)), list(dict)
    :param indentation: indent all by number of spaces
    :param index_of_colon: use to position : at the name string (-1 for no
        space)
    :param show_multi_line_character: option to hide "|"
    :returns: yaml formatted string array of name, value pair
    """
    strings = []
    if isinstance(value, list):
        # special case for single item lists:
        if len(value) == 1 and isinstance(value[0], str):
            # value = ["string"] should not be converted to
            # name:
            #   - string
            # but to "name: string" instead
            value = value[0]
        elif len(value) == 1 and isinstance(value[0], list) \
                and len(value[0]) == 1 and isinstance(value[0][0], str):
            # same applies to value = [["string"]]
            value = value[0][0]
    if isinstance(value, str):
        strings.append("{}{}{}: {}".format(
            ' ' * indentation, name, ' ' * (index_of_colon-len(name)),
            indent_multiline_string(value, indentation+4,
                                    show_multi_line_character)))
    elif isinstance(value, list):
        strings.append("{}{}{}: ".format(
            ' ' * indentation, name, ' ' * (index_of_colon-len(name))))
        for outer in value:
            # special case for single item sublists
            if isinstance(outer, list) and len(outer) == 1 \
                    and isinstance(outer[0], str):
                # outer = ["string"] should not be converted to
                # -
                #   - string
                # but to "- string" instead
                outer = outer[0]
            if isinstance(outer, str):
                strings.append("{}- {}".format(
                    ' ' * (indentation+4), indent_multiline_string(
                        outer, indentation+8, show_multi_line_character)))
            elif isinstance(outer, list):
                strings.append("{}- ".format(' ' * (indentation+4)))
                for inner in outer:
                    if isinstance(inner, str):
                        strings.append("{}- {}".format(
                            ' ' * (indentation+8), indent_multiline_string(
                                inner, indentation+12,
                                show_multi_line_character)))
            elif isinstance(outer, dict):
                # ABLABEL'd lists
                for k in outer:
                    strings += convert_to_yaml(
                        "- " + k, outer[k], indentation+4, index_of_colon,
                        show_multi_line_character)
    return strings


def indent_multiline_string(input: Union[str, List], indentation: int,
                            show_multi_line_character: bool) -> str:
    # if input is a list, convert to string first
    if isinstance(input, list):
        input = list_to_string(input, "")
    # format multiline string
    if "\n" in input or ": " in input:
        lines = ["|"] if show_multi_line_character else [""]
        for line in input.split("\n"):
            lines.append("{}{}".format(' ' * indentation, line.strip()))
        return '\n'.join(lines)
    return input.strip()


def get_new_contact_template(
        supported_private_objects: Optional[List[str]] = None) -> str:
    formatted_private_objects = []
    if supported_private_objects:
        formatted_private_objects.append("")
        longest_key = max(supported_private_objects, key=len)
        for object in supported_private_objects:
            formatted_private_objects += convert_to_yaml(
                object, "", 12, len(longest_key)+1, True)
    template = pathlib.Path(__file__).parent.parent / 'data' / 'template.yaml'
    with template.open() as temp:
        return temp.read().format('\n'.join(formatted_private_objects))



"""Tests for the user interaction functions."""

import unittest
from unittest import mock

from khard.helpers import interactive

from .helpers import mock_stream


class Select(unittest.TestCase):

    def _test(expected, include_none=None):
        input_list = ["a", "b", "c"]
        if include_none is None:
            return interactive.select(input_list)
        else:
            return interactive.select(input_list, include_none)

    def test_selection_index_is_1_based(self):
        with mock.patch("builtins.input", lambda x: "1"):
            actual = self._test()
        self.assertEqual(actual, "a")

    def test_typing_a_non_number_prints_a_message_and_repeats(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["foo", "2"])):
            with mock_stream() as stdout:
                actual = self._test()
        stdout = stdout.getvalue()
        self.assertEqual(stdout, "Please enter an index value between 1 and 3 "
                         "or q to quit.\n")
        self.assertEqual(actual, "b")

    def test_out_of_bounds_repeats(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["5", "2"])):
            with mock_stream() as stdout:
                actual = self._test()
        stdout = stdout.getvalue()
        self.assertEqual(stdout, "Please enter an index value between 1 and 3 "
                         "or q to quit.\n")
        self.assertEqual(actual, "b")


class Confirm(unittest.TestCase):

    def test_y_is_true(self):
        with mock.patch("builtins.input", lambda x: "y"):
            self.assertTrue(interactive.confirm(""))

    def test_n_is_false(self):
        with mock.patch("builtins.input", lambda x: "n"):
            self.assertFalse(interactive.confirm(""))

    def test_Y_is_true(self):
        with mock.patch("builtins.input", lambda x: "Y"):
            self.assertTrue(interactive.confirm(""))

    def test_N_is_false(self):
        with mock.patch("builtins.input", lambda x: "N"):
            self.assertFalse(interactive.confirm(""))

    def test_full_word_yes_is_not_accepted(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["yes", "n"])):
            with mock_stream():
                self.assertFalse(interactive.confirm(""))

    def test_full_word_no_is_not_accepted(self):
        with mock.patch("builtins.input", mock.Mock(side_effect=["no", "y"])):
            with mock_stream():
                self.assertTrue(interactive.confirm(""))



"""Tests for runtime type conversions"""

import datetime
import unittest

from khard.helpers.typing import convert_to_vcard, list_to_string, \
    ObjectType, string_to_date


class ConvertToVcard(unittest.TestCase):

    def test_returns_strings(self):
        value = "some text"
        actual = convert_to_vcard("test", value, ObjectType.str)
        self.assertEqual(value, actual)

    def test_returns_lists(self):
        value = ["some", "text"]
        actual = convert_to_vcard("test", value, ObjectType.list)
        self.assertListEqual(value, actual)

    def test_fail_if_not_string(self):
        value = ["some", "text"]
        with self.assertRaises(ValueError):
            convert_to_vcard("test", value, ObjectType.str)

    def test_upgrades_string_to_list(self):
        value = "some text"
        actual = convert_to_vcard("test", value, ObjectType.list)
        self.assertListEqual([value], actual)

    def test_fails_if_string_lists_are_not_homogenous(self):
        value = ["some", ["nested", "list"]]
        with self.assertRaises(ValueError):
            convert_to_vcard("test", value, ObjectType.list)

    def test_empty_list_items_are_filtered(self):
        value = ["some", "", "text", "", "more text"]
        actual = convert_to_vcard("test", value, ObjectType.list)
        self.assertListEqual(["some", "text", "more text"], actual)

    def test_strings_are_stripped(self):
        value = " some text "
        actual = convert_to_vcard("test", value, ObjectType.str)
        self.assertEqual("some text", actual)

    def test_strings_in_lists_are_stripped(self):
        value = [" some ", " text "]
        actual = convert_to_vcard("test", value, ObjectType.list)
        self.assertListEqual(["some", "text"], actual)


class ListToString(unittest.TestCase):

    def test_empty_list_returns_empty_string(self):
        the_list = []
        delimiter = ' '
        expected = ''
        actual = list_to_string(the_list, delimiter)
        self.assertEqual(actual, expected)

    def test_simple_list(self):
        the_list = ['a', 'bc', 'def']
        delimiter = ' '
        expected = 'a bc def'
        actual = list_to_string(the_list, delimiter)
        self.assertEqual(actual, expected)

    def test_simple_nested_list(self):
        the_list = ['a', 'bc', ['x', 'y', 'z'], 'def']
        delimiter = ' '
        expected = 'a bc x y z def'
        actual = list_to_string(the_list, delimiter)
        self.assertEqual(actual, expected)

    def test_multi_level_nested_list(self):
        the_list = ['a', ['b', ['c', [[['x', 'y']]]]], 'z']
        delimiter = ' '
        expected = 'a b c x y z'
        actual = list_to_string(the_list, delimiter)
        self.assertEqual(actual, expected)

    def test_list_to_string_passes_through_other_objects(self):
        self.assertIs(list_to_string(None, "foo"), None)
        self.assertIs(list_to_string(42, "foo"), 42)
        self.assertIs(list_to_string("foo bar", "foo"), "foo bar")


class StringToDate(unittest.TestCase):

    date = datetime.datetime(year=1900, month=1, day=2)
    time = datetime.datetime(year=1900, month=1, day=2, hour=12, minute=42,
                             second=17)
    zone = datetime.datetime(year=1900, month=1, day=2, hour=12, minute=42,
                             second=17, tzinfo=datetime.timezone.utc)

    def test_mmdd_format(self):
        string = '--0102'
        result = string_to_date(string)
        self.assertEqual(result, self.date)

    def test_mm_dd_format(self):
        string = '--01-02'
        result = string_to_date(string)
        self.assertEqual(result, self.date)

    def test_yyyymmdd_format(self):
        string = '19000102'
        result = string_to_date(string)
        self.assertEqual(result, self.date)

    def test_yyyy_mm_dd_format(self):
        string = '1900-01-02'
        result = string_to_date(string)
        self.assertEqual(result, self.date)

    def test_yyyymmddThhmmss_format(self):
        string = '19000102T124217'
        result = string_to_date(string)
        self.assertEqual(result, self.time)

    def test_yyyy_mm_ddThh_mm_ss_format(self):
        string = '1900-01-02T12:42:17'
        result = string_to_date(string)
        self.assertEqual(result, self.time)

    def test_yyyymmddThhmmssZ_format(self):
        string = '19000102T124217Z'
        result = string_to_date(string)
        self.assertEqual(result, self.time)

    def test_yyyy_mm_ddThh_mm_ssZ_format(self):
        string = '1900-01-02T12:42:17Z'
        result = string_to_date(string)
        self.assertEqual(result, self.time)

    def test_yyyymmddThhmmssz_format(self):
        string = '19000102T064217-06:00'
        result = string_to_date(string)
        self.assertEqual(result, self.zone)

    def test_yyyy_mm_ddThh_mm_ssz_format(self):
        string = '1900-01-02T06:42:17-06:00'
        result = string_to_date(string)
        self.assertEqual(result, self.zone)



"""Tests for the custom YAML format."""
# pylint: disable=missing-docstring

import datetime
from io import StringIO
import unittest
from unittest import mock
import copy

from ruamel.yaml import YAML

from khard.carddav_object import CarddavObject
import khard.helpers

from .helpers import TestYAMLEditable as create_test_card


def to_yaml(data):
    if 'First name' not in data:
        data['First name'] = 'Nobody'
    stream = StringIO()
    YAML().dump(data, stream)
    return stream.getvalue()


def parse_yaml(yaml=''):
    """Parse some yaml string into a CarddavObject

    :param yaml: the yaml input string to parse
    :type yaml: str
    :returns: the parsed CarddavObject
    :rtype: CarddavObject
    """
    return CarddavObject.from_yaml(address_book=mock.Mock(path='foo-path'),
                                   yaml=yaml, supported_private_objects=[],
                                   version='3.0', localize_dates=False)


class EmptyFieldsAndSpaces(unittest.TestCase):

    def test_empty_birthday_in_yaml_input(self):
        empty_birthday = "First name: foo\nBirthday:"
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(empty_birthday)
        self.assertIsNone(x.birthday)

    def test_only_spaces_in_birthday_in_yaml_input(self):
        spaces_birthday = "First name: foo\nBirthday:  "
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(spaces_birthday)
        self.assertIsNone(x.birthday)

    def test_empty_anniversary_in_yaml_input(self):
        empty_anniversary = "First name: foo\nAnniversary:"
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(empty_anniversary)
        self.assertIsNone(x.anniversary)

    def test_empty_organisation_in_yaml_input(self):
        empty_organisation = "First name: foo\nOrganisation:"
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(empty_organisation)
        self.assertListEqual(x.organisations, [])

    def test_empty_nickname_in_yaml_input(self):
        empty_nickname = "First name: foo\nNickname:"
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(empty_nickname)
        self.assertListEqual(x.nicknames, [])

    def test_empty_role_in_yaml_input(self):
        empty_role = "First name: foo\nRole:"
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(empty_role)
        self.assertListEqual(x.roles, [])

    def test_empty_title_in_yaml_input(self):
        empty_title = "First name: foo\nTitle:"
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(empty_title)
        self.assertListEqual(x.titles, [])

    def test_empty_categories_in_yaml_input(self):
        empty_categories = "First name: foo\nCategories:"
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(empty_categories)
        self.assertListEqual(x.categories, [])

    def test_empty_webpage_in_yaml_input(self):
        empty_webpage = "First name: foo\nWebpage:"
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(empty_webpage)
        self.assertListEqual(x.webpages, [])

    def test_empty_note_in_yaml_input(self):
        empty_note = "First name: foo\nNote:"
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(empty_note)
        self.assertListEqual(x.notes, [])


class yaml_ablabel(unittest.TestCase):

    def test_ablabelled_url_in_yaml_input(self):
        ablabel_url = "First name: foo\nWebpage:\n - http://example.com\n" \
                      " - github: https://github.com/scheibler/khard"
        with mock.patch("khard.carddav_object.logger"):
            x = parse_yaml(ablabel_url)
        self.assertListEqual(x.webpages, [
            'http://example.com',
            {'github': 'https://github.com/scheibler/khard'}])


class UpdateVcardWithYamlUserInput(unittest.TestCase):

    _date = datetime.datetime(2000, 1, 1)
    _datetime = datetime.datetime(2013, 4, 2, 13, 14, 15)
    _no_year = datetime.datetime(1900, 1, 1)

    def test_update_org_simple(self):
        card = create_test_card()
        data = {'Organisation': 'Foo'}
        data = to_yaml(data)
        card.update(data)
        self.assertListEqual(card.organisations, [['Foo']])

    def test_update_org_multi(self):
        card = create_test_card()
        orgs = ['foo', 'bar', 'baz']
        data = {'Organisation': orgs}
        data = to_yaml(data)
        card.update(data)
        self.assertListEqual(card.organisations, sorted([[x] for x in orgs]))

    def test_update_org_complex(self):
        card = create_test_card()
        org = ['org.', 'dep.', 'office']
        data = {'Organisation': [org]}
        data = to_yaml(data)
        card.update(data)
        self.assertListEqual(card.organisations, [org])

    def test_update_categories_simple(self):
        card = create_test_card()
        data = {'Categories': 'foo'}
        data = to_yaml(data)
        card.update(data)
        self.assertListEqual(card.categories, ['foo'])

    def test_update_categories_multi(self):
        card = create_test_card()
        cat = ['foo', 'bar', 'baz']
        data = {'Categories': cat}
        data = to_yaml(data)
        card.update(data)
        self.assertListEqual(card.categories, cat)

    def test_update_bday_date(self):
        card = create_test_card()
        data = {'Birthday': '2000-01-01'}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.birthday, self._date)

    def test_update_bday_without_year(self):
        card = create_test_card(version="4.0")
        data = {'Birthday': '--01-01'}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.birthday, self._no_year)

    def test_update_bday_with_text(self):
        card = create_test_card(version="4.0")
        data = {'Birthday': 'text= some day maybe'}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.birthday, 'some day maybe')

    def test_update_bday_with_date_and_time(self):
        card = create_test_card()
        data = {'Birthday': '2013-04-02T13:14:15'}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.birthday, self._datetime)

    def test_update_anniverary(self):
        card = create_test_card()
        data = {'Anniversary': '2000-01-01'}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.anniversary, self._date)

    def test_update_anniversary_without_year(self):
        card = create_test_card(version="4.0")
        data = {'Anniversary': '--01-01'}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.anniversary, self._no_year)

    def test_update_anniversary_with_text(self):
        card = create_test_card(version="4.0")
        data = {'Anniversary': 'text= some day maybe'}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.anniversary, 'some day maybe')

    def test_update_anniversary_with_date_and_time(self):
        card = create_test_card()
        data = {'Anniversary': '2013-04-02T13:14:15'}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.anniversary, self._datetime)

    def test_update_name_simple(self):
        card = create_test_card()
        data = {'First name': 'first', 'Last name': 'last'}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.get_first_name_last_name(), 'first last')

    def test_update_fn(self):
        card = create_test_card()
        fn = 'me myself and i'
        data = {'Formatted name': fn}
        data = to_yaml(data)
        card.update(data)
        self.assertEqual(card.formatted_name, fn)

    def test_parse_field(self):
        """Test round-trip of a field to/from YAML"""
        card = create_test_card()
        data = "First name: Nobody\n"
        data += "\n".join(khard.helpers.convert_to_yaml("Note", "foobar", 0,
                                                        5, True))
        card.update(data)
        self.assertListEqual(card.notes, ["foobar"])

    def test_parse_field_with_colon(self):
        """Test round-trip of a field containing ': ' to/from YAML"""
        card = create_test_card()
        data = "First name: Nobody\n"
        data += "\n".join(khard.helpers.convert_to_yaml("Note", "foo: bar", 0,
                                                        5, True))
        card.update(data)
        self.assertListEqual(card.notes, ["foo: bar"])

    def test_vcard_round_trip(self):
        """Test a VCARD can be converted to YAML and back unchanged"""
        card = create_test_card()
        card._add_organisation("ACME, Inc")
        card._add_note("foo: bar")
        card2 = copy.copy(card)
        yaml = card.to_yaml()
        card.update(yaml)
        self.assertEqual(card.vcard.serialize(), card2.vcard.serialize())



"""Tests for the helpers module."""
# pylint: disable=missing-docstring

import unittest

from khard import helpers


class ConvertToYAML(unittest.TestCase):

    def test_colon_handling(self):
        result = helpers.convert_to_yaml("Note", "foo: bar", 0, 5, True)
        self.assertListEqual(result, ["Note : |\n    foo: bar"])

    def test_none_values_produce_no_output(self):
        result = helpers.convert_to_yaml("Note", None, 0, 5, True)
        self.assertListEqual(result, [])

    def test_empty_strings_produce_empty_values(self):
        result = helpers.convert_to_yaml("Note", "", 0, 5, True)
        self.assertListEqual(result, ["Note : "])


if __name__ == "__main__":
    unittest.main()
